"""Project: Eskapade - A python-based package for data analysis.

Class: HistogramFillerBase

Created: 2017/03/21

Description:
    Algorithm to fill histogrammar sparse-bin histograms.
    It is possible to do cleaning of these histograms by
    rejecting certain keys or removing inconsistent data types.
    Timestamp columns are converted to nanoseconds before
    the binning is applied.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import numpy as np
import pandas as pd

from eskapade import process_manager, DataStore, Link, StatusCode

# numeric datatypes get converted to an index, which is then used for value counting
NUMERIC_SUBSTR = [np.dtype('int'), np.dtype('float'), np.dtype('double')]

# string datatype get treated as categories
STRING_SUBSTR = [np.dtype('str'), np.dtype('object'), np.dtype('bool')]

# timestamps are converted to nanoseconds (int)
TIME_SUBSTR = [np.dtype('datetime64[ns]'), np.datetime64, np.dtype('<M8')]
NUM_NS_DAY = 24 * 3600 * int(1e9)


class HistogramFillerBase(Link):
    """Base class link to fill histograms.

    It is possible to do after-filling cleaning of these histograms by rejecting certain
    keys or removing inconsistent data types. Timestamp columns are
    converted to nanoseconds before the binning is applied. Final histograms
    are stored in the datastore.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        Store and do basic check on the attributes of link HistogramFillerBase.

        :param str name: name of link
        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store histograms in data store
        :param list columns: colums to pick up from input data. (default is all columns)
        :param dict bin_specs: dictionaries used for rebinning numeric or timestamp columns

        Example bin_specs dictionary is:

        >>> bin_specs = {'x': {'bin_width': 1, 'bin_offset': 0},
                         'y': {'bin_edges': [0, 2, 3, 4, 5, 7, 8]}}

        :param dict var_dtype: dict of datatypes of the columns to study from dataframe.
                               If not provided, try to determine datatypes directy from dataframe.
        :param bool store_at_finalize: Store histograms in datastore at finalize(), not at
                                       execute(). Useful when looping over datasets. Default is False.
        :param drop_keys dict: dictionary used for dropping specific keys from bins dictionaries of histograms

        Example drop_keys dictionary is:

        >>> drop_keys = {'x': [1,4,8,19],
                         'y': ['apple', 'pear', 'tomato'],
                         'x:y': [(1, 'apple'), (19, 'tomato')]}
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'HistogramFillerBase'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs,
                             read_key=None,
                             store_key=None,
                             columns=[],
                             bin_specs={},
                             var_dtype={},
                             drop_keys={},
                             store_at_finalize=False)

        self._unit_bin_specs = {'bin_width': 1.0, 'bin_offset': 0.0}
        self._unit_timestamp_specs = {'bin_width': pd.Timedelta(days=30).value,
                                      'bin_offset': pd.Timestamp('2010-01-04').value}

        # these get filled during execution
        self._hists = {}

        # initialize attributes
        self.all_columns = []
        self.str_cols = []
        self.num_cols = []
        self.dt_cols = []

    def initialize(self):
        """Initialize the link."""
        # check basic attribute settings
        assert isinstance(self.read_key, str) and len(self.read_key), 'read_key has not been set correctly'
        if self.store_key is not None:
            assert isinstance(self.store_key, str) and len(self.store_key), 'store_key has not been set to string'

        # default histogram creation is at execute(). Storage at finalize is useful for
        # looping over datasets.
        if self.store_at_finalize:
            self.logger.debug('Storing (and possible post-processing) at finalize, not execute.')

        # check that columns are set correctly.
        for i, c in enumerate(self.columns):
            if isinstance(c, str):
                self.columns[i] = [c]
            elif not isinstance(self.columns[i], list):
                raise TypeError('Columns "{}" needs to be a string or list of strings'.format(self.columns[i]))

        # check for supported data types
        for k in self.var_dtype:
            try:
                self.var_dtype[k] = np.dtype(self.var_dtype[k]).type
                if self.var_dtype[k] is np.string_ or self.var_dtype[k] is np.object_:
                    self.var_dtype[k] = str
            except BaseException:
                raise RuntimeError('unknown assigned datatype to variable "{}"'.format(k))

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Execute() four things:

        * check presence and data type of requested columns
        * timestamp variables are converted to nanosec (integers)
        * do the actual value counting based on categories and created indices
        * then convert to histograms and add to datastore
        """
        ds = process_manager.service(DataStore)

        # basic checks on contensts of the data frame
        if self.read_key not in ds:
            raise KeyError('key "{}" not in data store'.format(self.read_key))
        df = ds[self.read_key]
        self.assert_dataframe(df)

        # determine all possible columns, used for comparison below
        self.all_columns = self.get_all_columns(df)

        # copy all columns from the dataframe?
        if not self.columns:
            self.columns = self.all_columns
            for i, c in enumerate(self.all_columns):
                self.columns[i] = [c]

        # 1. check presence and data type of requested columns
        # sort columns into numerical, timestamp and category based
        self.categorize_columns(df)

        # 2. timestamp variables are converted to ns here
        idf = self.process_columns(df)

        # 3. do the actual histogram/counter filling
        for c in self.columns:
            name = ':'.join(c)
            self.logger.debug('Processing column(s) "{col}".', col=name)
            self.fill_histogram(idf, c)

        # cleanup temp df
        del idf

        # 4. storage
        if not self.store_at_finalize:
            self.process_and_store()

        return StatusCode.Success

    def finalize(self):
        """Finalize the link.

        Store Histograms here, if requested.
        """
        # convert to histograms and add to datastore
        if self.store_at_finalize:
            self.process_and_store()

        return StatusCode.Success

    def process_and_store(self):
        """Store (and possibly process) histogram objects."""
        ds = process_manager.service(DataStore)

        if self.store_key is not None:
            ds[self.store_key] = self._hists

    def assert_dataframe(self, df):
        """Check that input data is a filled pandas data frame.

        :param df: input (pandas) data frame
        """
        if not isinstance(df, pd.DataFrame):
            raise TypeError('retrieved object not of type pandas DataFrame')
        assert len(df.index) > 0, 'input dataframe is empty'

    def get_all_columns(self, data):
        """Retrieve all columns / keys from input data.

        :param data: input data sample (pandas dataframe or dict)
        :returns: list of columns
        :rtype: list
        """
        if isinstance(data, pd.DataFrame):
            all_columns = sorted(data.columns.tolist())
        else:
            raise RuntimeError('Cannot determine columns in input data found for {!s}'.format(self))

        return all_columns

    def get_data_type(self, df, col):
        """Get data type of dataframe column.

        :param df: input data frame
        :param str col: column
        """
        if col not in df.columns:
            raise KeyError('column "{0:s}" not in input dataframe'.format(col))
        return df[col].dtype

    def categorize_columns(self, df):
        """Categorize columns of dataframe by data type.

        :param df: input (pandas) data frame
        """
        # check presence and data type of requested columns
        # sort columns into numerical, timestamp and category based
        for c in self.columns:
            for col in c:
                if col not in df.columns:
                    raise KeyError('Column "{0:s}" not in dataframe "{1:s}".'.format(col, self.read_key))
                dt = self.get_data_type(df, col)
                if col not in self.var_dtype:
                    self.var_dtype[col] = dt.type
                    if (self.var_dtype[col] is np.string_) or (self.var_dtype[col] is np.object_):
                        self.var_dtype[col] = str
                if not any(dt in types for types in (STRING_SUBSTR, NUMERIC_SUBSTR, TIME_SUBSTR)):
                    raise TypeError('Cannot process column "{0:s}" of data type "{1!s}".'.format(col, dt))
                is_number = isinstance(dt.type(), np.number)
                is_timestamp = isinstance(dt.type(), np.datetime64)
                colset = self.num_cols if is_number else self.dt_cols if is_timestamp else self.str_cols
                if col not in colset:
                    colset.append(col)
                self.logger.debug('Data type of column "{col}" is "{type}".', col=col, type=self.var_dtype[col])

    def process_columns(self, df):
        """Process columns before histogram filling.

        Specifically, convert timestamp columns to integers

        :param df: input (pandas) data frame
        :returns: output (pandas) data frame with converted timestamp columns
        :rtype: pandas DataFrame
        """
        # timestamp variables are converted to ns here
        # make temp df for value counting (used below)
        idf = df[self.num_cols + self.str_cols].copy(deep=False)
        for col in self.dt_cols:
            self.logger.debug('Converting column "{col}" of type "{type}" to nanosec.',
                              col=col, type=self.var_dtype[col])
            idf[col] = df[col].apply(to_ns)
        return idf

    def fill_histogram(self, idf, c):
        """Fill input histogram with column(s) of input dataframe.

        :param idf: input data frame used for filling histogram
        :param list c: histogram column(s)
        """
        return

    def drop_requested_keys(self, name, counts):
        """Drop requested keys from counts dictionary.

        :param string name: key of drop_keys dict to get array of keys to be dropped
        :param dict counts: counts dictionary to drop specific keys from
        :returns: count dict without dropped keys
        """
        # drop requested keys
        if name in self.drop_keys:
            keys_to_drop = self.drop_keys[name]
            if not isinstance(keys_to_drop, list):
                raise TypeError('drop_keys value needs to be a list of values')
            for key in keys_to_drop:
                if key in counts:
                    self.logger.debug('Removing key "{key}" with value: "{value}", as requested.',
                                      key=key, value=counts[key])
                    del counts[key]
        return counts


def to_ns(x):
    """Convert input timestamps to nanoseconds (integers).

    :param x: value to be converted
    :returns: converted value
    :rtype: int
    """
    if pd.isnull(x):
        return 0
    try:
        return pd.to_datetime(x).value
    except Exception:
        if hasattr(x, '__str__'):
            return pd.to_datetime(str(x)).value
    return 0


def to_str(val):
    """Convert input to (array of) string(s).

    :param val: value to be converted
    :returns: converted value
    :rtype: str or np.ndarray
    """
    if isinstance(val, str):
        return val
    elif hasattr(val, '__iter__'):
        return np.asarray(list(map(lambda s: s if isinstance(s, str) else str(s) if hasattr(s, '__str__') else '',
                                   val)))
    elif hasattr(val, '__str__'):
        return str(val)

    return ''


def only_str(val):
    """Pass input value or array only if it is a string.

    :param val: value to be evaluated
    :returns: evaluated value
    :rtype: str or np.ndarray
    """
    if isinstance(val, str):
        return val
    elif hasattr(val, '__iter__'):
        return np.asarray(list(filter(lambda s: isinstance(s, str), val)))

    return None


def only_bool(val):
    """Pass input value or array only if it is a bool.

    :param val: value to be evaluated
    :returns: evaluated value
    :rtype: np.bool or np.ndarray
    """
    if isinstance(val, (np.bool_, bool)):
        return np.bool(val)
    elif hasattr(val, '__iter__') and not isinstance(val, str):
        return np.asarray(list(filter(lambda s: isinstance(s, (np.bool_, bool)), val)))

    return None


def only_int(val):
    """Pass input val value or array only if it is an integer.

    :param val: value to be evaluated
    :returns: evaluated value
    :rtype: np.int64 or np.ndarray
    """
    if isinstance(val, (np.int64, int)):
        return np.int64(val)
    elif hasattr(val, '__iter__') and not isinstance(val, str):
        return np.asarray(list(filter(lambda s: isinstance(s, (np.int64, int)), val)))

    return None


def only_float(val):
    """Pass input val value or array only if it is a float.

    :param val: value to be evaluated
    :returns: evaluated value
    :rtype: np.float64 or np.ndarray
    """
    if isinstance(val, (np.float64, float)):
        return np.float64(val)
    elif hasattr(val, '__iter__') and not isinstance(val, str):
        return np.asarray(list(filter(lambda s: isinstance(s, (np.float64, float)), val)))

    return np.nan


QUANTITY = {str: only_str, np.str_: only_str,
            int: only_int, np.int64: only_int,
            bool: only_bool, np.bool_: only_bool,
            float: only_float, np.float64: only_float,
            np.datetime64: only_int}


def value_to_bin_index(val, **kwargs):
    """Convert value to bin index.

    Convert a numeric or timestamp column to an integer bin index.

    :param bin_width: bin_width value needed to convert column to an integer bin index
    :param bin_offset: bin_offset value needed to convert column to an integer bin index
    """
    try:
        # NOTE this notation also works for timestamps
        bin_width = kwargs.get('bin_width', 1)
        bin_offset = kwargs.get('bin_offset', 0)
        bin_index = int(np.floor((val - bin_offset) / bin_width))
        return bin_index
    except BaseException:
        pass
    return val


def value_to_bin_center(val, **kwargs):
    """Convert value to bin center.

    Convert a numeric or timestamp column to a common bin center value.

    :param bin_width: bin_width value needed to convert column to a common bin center value
    :param bin_offset: bin_offset value needed to convert column to a common bin center value
    """
    try:
        # NOTE this notation also works for timestamps, and does not change the
        # unit
        bin_width = kwargs.get('bin_width', 1)
        bin_offset = kwargs.get('bin_offset', 0)
        bin_index = int(np.floor((val - bin_offset) / bin_width))
        obj_type = type(bin_width)
        return bin_offset + obj_type((bin_index + 0.5) * bin_width)
    except BaseException:
        pass
    return val



"""Project: Eskapade - A python-based package for data analysis.

Class : DfSummary

Created: 2017/02/17

Description:
    Link to create a statistics summary of data frame columns
    or of a set of histograms.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import numpy as np
import pandas as pd
import tabulate

from eskapade import resources, visualization
from eskapade import process_manager, Link, DataStore, StatusCode
from eskapade.analysis import statistics
from eskapade.core import persistence

NUMBER_OF_BINS = 30


class DfSummary(Link):
    """Create a summary of a dataframe.

    Creates a report page for each variable in data frame, containing:

    * a profile of the column dataset
    * a nicely scaled plot of the column dataset

    Example 1 is available in: tutorials/esk301_dfsummary_plotter.py

    Example 2 is available in: tutorials/esk303_histogram_filling_plotting.py
    Empty histograms are automatically skipped from processing.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param str read_key: key of input dataframe (or histogram-dict) to read from data store
        :param str results_path: output path of summary result files
        :param list columns: columns (or histogram keys) pick up from input data to make & plot summaries for
        :param list hist_keys: alternative to columns (optional)
        :param dict var_labels: dict of column names with a label per column
        :param dict var_units: dict of column names with a unit per column
        :param dict var_bins: dict of column names with the number of bins per column. Default per column is 30.
        :param str hist_y_label: y-axis label to plot for all columns. Default is 'Bin Counts'.
        :param str pages_key: data store key of existing report pages
        """
        # initialize Link
        Link.__init__(self, kwargs.pop('name', 'df_summary'))

        # process keyword arguments
        self._process_kwargs(kwargs, read_key='', results_path='', columns=[],
                             hist_keys=[],
                             var_labels={}, var_units={}, var_bins={},
                             hist_y_label='Bin counts', pages_key='')
        self.check_extra_kwargs(kwargs)

        # initialize attributes
        self.pages = []
        self.nan_counts = []

    def _process_results_path(self):
        """Process results_path argument."""
        if not self.results_path:
            self.results_path = persistence.io_path('results_data', 'report')
        persistence.create_dir(self.results_path)

    def initialize(self):
        """Initialize the link."""
        # check input arguments
        self.check_arg_types(read_key=str, pages_key=str)
        self.check_arg_types(recurse=True, allow_none=True, columns=str, hist_keys=str, var_labels=str, var_units=str)
        self.check_arg_vals('read_key')

        # read report templates
        with open(resources.template('df_summary_report.tex')) as templ_file:
            self.report_template = templ_file.read()
        with open(resources.template('df_summary_report_page.tex')) as templ_file:
            self.page_template = templ_file.read()

        self._process_results_path()
        # add hist_keys to columns, ensure sum is a unique set
        self.columns += self.hist_keys
        self.columns = sorted(list(set(self.columns)))

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Creates a report page for each variable in data frame.

        * create statistics object for column
        * create overview table of column variable
        * plot histogram of column variable
        * store plot

        :returns: execution status code
        :rtype: StatusCode
        """
        ds = process_manager.service(DataStore)

        # fetch and check input data frame
        data = ds.get(self.read_key, None)
        if data is None:
            self.logger.fatal('No input data "{self.read_key}" found in data store for {self!s}.', self=self)
            raise RuntimeError('no input data found for {!s}'.format(self))
        else:
            self.assert_data_type(data)

        # create report page for histogram
        if self.pages_key:
            self.pages = ds.get(self.pages_key, [])
            if not isinstance(self.pages, list):
                raise TypeError('pages key "{}" does not refer to a list'.format(self.pages_key))

        # determine all possible columns, used for comparison below
        all_columns = self.get_all_columns(data)
        if not self.columns:
            self.columns = all_columns

        for name in self.columns[:]:
            # check if column is in data frame
            if name not in all_columns:
                self.logger.warning('Key "{key}" not in input data; skipping.', key=name)
                self.columns.remove(self.columns.index(name))
                continue
            self.logger.debug('Processing "{col}".', col=name)
            sample = self.get_sample(data, name)
            self.process_sample(name, sample)

        # add nan histogram to summary if present
        if self.nan_counts:
            nan_hist = self.nan_counts, self.columns
            self.process_nan_histogram(nan_hist, self.get_length(data))

        # storage
        if self.pages_key:
            ds[self.pages_key] = self.pages

        return StatusCode.Success

    def finalize(self):
        """Finalize the link."""
        # write report file
        with open('{}/report.tex'.format(self.results_path), 'w') as report_file:
            report_file.write(self.report_template.replace('INPUT_PAGES', ''.join(self.pages)))

        return StatusCode.Success

    def assert_data_type(self, data):
        """Check type of input data.

        :param data: input data sample (pandas dataframe or dict)
        """
        if not isinstance(data, pd.DataFrame) and not isinstance(data, dict):
            raise AssertionError('input data should be pandas dataframe or dict (with histograms)')

    def get_all_columns(self, data):
        """Retrieve all columns / keys from input data.

        :param data: input data sample (pandas dataframe or dict)
        :returns: list of columns
        :rtype: list
        """
        if isinstance(data, pd.DataFrame):
            all_columns = sorted(data.columns.tolist())
        elif isinstance(data, dict):
            # dict of histograms
            all_columns = sorted(data.keys())
        else:
            raise RuntimeError('cannot determine columns in input data found for {!s}'.format(self))

        return all_columns

    def get_sample(self, data, key):
        """Retrieve speficic column or item from input data.

        :param data: input data (pandas dataframe or dict)
        :param str key: column key
        :returns: data series or item
        """
        return data[key]

    def get_length(self, data):
        """Get length of data set.

        :param data: input data (pandas dataframe or dict)
        :returns: length of data set
        """
        return len(data)

    def process_sample(self, name, sample):
        """Process various possible data samples.

        :param str name: name of sample
        :param sample: input pandas series object or histogram
        """
        # process pandas series (plot and make summary table)
        if isinstance(sample, pd.core.series.Series):
            self.process_series(name, sample)
        # else process histograms
        elif hasattr(sample, 'n_dim'):
            if sample.n_dim == 1:
                self.process_1d_histogram(name, sample)
            elif sample.n_dim == 2:
                self.process_2d_histogram(name, sample)

    def process_series(self, col, sample):
        """Create statistics of and plot input pandas series.

        :param str col: name of the series
        :param sample: input pandas series object
        """
        # skip columns consisting entirely of nans
        nan_cnt = sample.isnull().sum()
        self.nan_counts.append(nan_cnt)
        if nan_cnt == len(sample.index):
            self.logger.debug('Column "{col}" consists of nans only; skipping.', col=col)
            return

        # 1. create statistics object for column
        var_label = self.var_labels.get(col, col)
        stats = statistics.ArrayStats(sample, col, unit=self.var_units.get(col, ''), label=var_label)
        # evaluate statitical properties of array
        stats.create_stats()

        # make histogram
        nphist = stats.make_histogram(var_bins=self.var_bins.get(col, NUMBER_OF_BINS))

        # determine histogram properties for plotting
        x_label = stats.get_x_label()
        y_label = self.hist_y_label if self.hist_y_label else None
        is_num = stats.get_col_props()['is_num']
        is_ts = stats.get_col_props()['is_ts']
        hist_file_name = 'hist_{}.pdf'.format(col)
        pdf_file_name = '{0:s}/{1:s}'.format(self.results_path, hist_file_name)

        # 3. plot histogram of column variable
        visualization.vis_utils.plot_histogram(nphist, x_label=x_label, y_label=y_label, is_num=is_num, is_ts=is_ts,
                                               pdf_file_name=pdf_file_name)

        # create overview table of column variable
        stats_table = stats.get_latex_table()

        # create page string for report
        self.pages.append(self.page_template.replace('VAR_LABEL', var_label).replace('VAR_STATS_TABLE', stats_table)
                          .replace('VAR_HISTOGRAM_PATH', hist_file_name))

    def process_1d_histogram(self, name, hist):
        """Create statistics of and plot input 1d histogram.

        :param str name: name of the histogram
        :param hist: input histogram object
        """
        # datatype properties
        datatype = hist.datatype
        col_props = statistics.get_col_props(datatype)
        is_num = col_props['is_num']
        is_ts = col_props['is_ts']

        # skip empty histograms
        n_bins = hist.n_bins
        if n_bins == 0:
            self.logger.warning('Histogram "{name}" is empty; skipping.', name=name)
            return

        bin_labels = hist.bin_centers() if is_num else hist.bin_labels()
        bin_counts = hist.bin_entries()
        bin_edges = hist.bin_edges() if is_num else None

        if is_ts:
            to_timestamp = np.vectorize(lambda x: pd.Timestamp(x))
            bin_labels = to_timestamp(bin_labels)
            bin_edges = to_timestamp(bin_edges)

        # create statistics object for histogram
        var_label = self.var_labels.get(name, name)
        stats = statistics.ArrayStats(bin_labels, name, weights=bin_counts, unit=self.var_units.get(name, ''),
                                      label=var_label)
        # evaluate statitical properties of array
        stats.create_stats()

        # make nice plots here ...
        # for numbers and timestamps, make cropped histogram, between percentiles 5-95%
        # ... and project on existing binning.
        # for categories, accept top N number of categories in bins.
        # NB: bin_edges overrules var_bins (if it is not none)
        nphist = stats.make_histogram(var_bins=self.var_bins.get(name, NUMBER_OF_BINS), bin_edges=bin_edges)

        # determine histogram properties for plotting below
        x_label = stats.get_x_label()
        y_label = self.hist_y_label if self.hist_y_label else None
        hist_file_name = 'hist_{}.pdf'.format(name.replace(' ', '_'))
        pdf_file_name = '{0:s}/{1:s}'.format(self.results_path, hist_file_name)

        # matplotlib plot of histogram
        visualization.vis_utils.plot_histogram(nphist, x_label=x_label, y_label=y_label, is_num=is_num, is_ts=is_ts,
                                               pdf_file_name=pdf_file_name)

        # create overview table of histogram statistics
        stats_table = stats.get_latex_table()

        # create page string
        page_templ = self.page_template
        for kv in [('VAR_LABEL', var_label), ('VAR_STATS_TABLE', stats_table), ('VAR_HISTOGRAM_PATH', hist_file_name)]:
            page_templ = page_templ.replace(*kv)
        self.pages.append(page_templ)

    def process_2d_histogram(self, name, hist):
        """Create statistics of and plot input 2d histogram.

        :param str name: name of the histogram
        :param hist: input histogram object
        """
        if not hasattr(hist, 'xy_ranges_grid'):
            self.logger.warning('No plot for 2d hist "{name}"; cannot extract binning and values.', name=name)
            return
        try:
            nphist = hist.xy_ranges_grid()
        except BaseException:
            raise RuntimeError('Cannot extract binning and values from input histogram.')

        # calc some basic histogram statistics
        sum_entries = 0
        try:
            grid = nphist[2]
            ynum = len(grid)
            xnum = len(grid[0])
            for i in range(xnum):
                for j in range(ynum):
                    sum_entries += grid[i, j]
        except BaseException:
            pass
        sum_entries = int(sum_entries)

        # create LaTeX string of statistics
        table = [('count', '{:d}'.format(sum_entries))]
        stats_table = tabulate.tabulate(table, tablefmt='latex')

        # histogram attributes for plotting
        var_label = self.var_labels.get(name, name.replace(':', '_vs_'))
        try:
            xlab = name.split(':')[0]
            ylab = name.split(':')[1]
        except BaseException:
            xlab = 'unknown x'
            ylab = 'unknown y'
        hist_file_name = 'hist_{}.pdf'.format(name.replace(':', '_vs_').replace(' ', '_'))
        pdf_file_name = '{0:s}/{1:s}'.format(self.results_path, hist_file_name)

        # plot the 2d histogram
        visualization.vis_utils.plot_2d_histogram(nphist, x_lim=hist.x_lim(), y_lim=hist.y_lim(), title=var_label,
                                                  x_label=xlab, y_label=ylab, pdf_file_name=pdf_file_name)

        # create page string for report
        page_templ = self.page_template
        for kv in [('VAR_LABEL', var_label), ('VAR_STATS_TABLE', stats_table), ('VAR_HISTOGRAM_PATH', hist_file_name)]:
            page_templ = page_templ.replace(*kv)
        self.pages.append(page_templ)

    def process_nan_histogram(self, nphist, n_data):
        """Process nans histogram.

        Add nans histogram to pdf list

        :param nphist: numpy-style input histogram, consisting of comma-separaged bin_entries, bin_edges
        :param int n_data: number of entries in the processed data set
        """
        var_label = 'NaN count'
        x_label = 'Column name'
        y_label = self.hist_y_label if self.hist_y_label else None
        hist_file_name = 'hist_NaNs.pdf'
        pdf_file_name = '{0:s}/{1:s}'.format(self.results_path, hist_file_name)
        visualization.vis_utils.plot_histogram(nphist, x_label=x_label, y_label=y_label, is_num=False, is_ts=False,
                                               pdf_file_name=pdf_file_name)
        table = [('count', '{:d}'.format(n_data))]
        stats_table = tabulate.tabulate(table, tablefmt='latex')
        self.pages.append(self.page_template.replace('VAR_LABEL', var_label).replace('VAR_STATS_TABLE', stats_table)
                          .replace('VAR_HISTOGRAM_PATH', hist_file_name))



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk306_concatenate_reports

Created: 2017/03/28

Description:
    This macro illustrates how to concatenate the reports of several
    visualization links into one big report.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, resources, Chain
from eskapade import analysis, visualization
from eskapade import process_manager
from eskapade.logger import Logger, LogLevel

logger = Logger()

logger.debug('Now parsing configuration file esk306_concatenate_reports.')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk306_concatenate_reports'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

input_files = resources.fixture('correlated_data.sv.gz')

#########################################################################################
# --- now set up the chains and links based on configuration flags

data = Chain('Data')

# --- 0. readdata keeps on opening the next file in the file list.
#     all kwargs are passed on to pandas file reader.
read_data = analysis.ReadToDf(name='dflooper', key='accounts', reader='csv', sep=' ')
read_data.path = input_files
data.add(read_data)

# --- 1. add data-frame summary link to "Summary" chain
summarizer = visualization.DfSummary(name='Create_stats_overview',
                                     read_key=read_data.key, pages_key='report_pages')
data.add(summarizer)

# --- 2. Fill 2d histogrammar histograms
hf = analysis.HistogrammarFiller()
hf.read_key = 'accounts'
hf.store_key = 'hist'
hf.logger.log_level = LogLevel.DEBUG
hf.columns = [
    ['x1', 'x2'],
    ['x1', 'x3'],
    ['x1', 'x4'],
    ['x1', 'x5'],
    ['x2', 'x3'],
    ['x2', 'x4'],
    ['x2', 'x5'],
    ['x3', 'x4'],
    ['x3', 'x4'],
    ['x4', 'x5']]
hf._unit_bin_specs = {'bin_width': 0.2, 'bin_offset': 0.0}
data.add(hf)

hs = visualization.DfSummary(name='HistogramSummary1', read_key=hf.store_key, pages_key='report_pages')
data.add(hs)

# --- 3. make visualizations of correlations
corr_link = visualization.CorrelationSummary(name='correlation_summary',
                                             read_key=read_data.key, pages_key='report_pages')
data.add(corr_link)

#########################################################################################

logger.debug('Done parsing configuration file esk306_concatenate_reports.')



"""Project: Eskapade - A python-based package for data analysis.

Class: AssertInDs

Created: 2016/11/08

Description:
    Algorithm that asserts that items exists in the datastore

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager


class AssertInDs(Link):
    """Asserts that specified item(s) exists in the datastore."""

    def __init__(self, **kwargs):
        """Initialize link instance.

        Store the configuration of link AssertInDs

        :param str name: name of link
        :param lst keySet: list of keys to check
        """
        Link.__init__(self, kwargs.pop('name', 'AssertInDs'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs, keySet=[])
        self.check_extra_kwargs(kwargs)

    def execute(self):
        """Execute the link."""
        ds = process_manager.service(DataStore)

        for key in self.keySet:
            assert key in ds, 'Key {} not in DataStore.'.format(key)

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017/03/31

Description:
    Eskapade exceptions

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""


class MissingPackageError(Exception):
    """Exception raised if third-party package is missing."""

    def __init__(self, message='', required_by=''):
        """Set missing-package arguments.

        :param str message: message to show when raised
        :param str required_by: info on component that requires the package
        """
        mess = message if message else 'missing third-party package'
        if required_by:
            mess += ' (required by "{}")'.format(required_by)
        Exception.__init__(self, mess)



"""Project: Eskapade - A python-based package for data analysis.

Class: DsObjectDeleter

Created: 2016/11/08

Description:
    Algorithm to delete objects from the datastore.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import process_manager, ConfigObject, DataStore, Link, StatusCode


class DsObjectDeleter(Link):
    """Delete objects from the datastore.

    Delete objects from the DataStore by the key they are under, or keeps
    only the data by the specified keys.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param list deletion_keys: keys to clear. Overwrites clear_all to false.
        :param list deletion_classes: delete object(s) by class type.
        :param lsst keep_only: keys to keep. Overwrites clear_all to false.
        :param bool clear_all: clear all key-value pairs in the datastore. Default is true.
        """
        Link.__init__(self, kwargs.pop('name', 'DsObjectDeleter'))

        # process keyword arguments
        self._process_kwargs(kwargs, deletion_keys=[], deletion_classes=[], keep_only=[], clear_all=True)
        self.check_extra_kwargs(kwargs)

    def initialize(self):
        """Initialize the link."""
        # Overwrites clear_all to false if individual keys are set.
        if len(self.deletion_keys):
            self.clear_all = False
        if len(self.deletion_classes):
            self.clear_all = False
        if len(self.keep_only):
            self.clear_all = False

        return StatusCode.Success

    def execute(self):
        """Execute the link."""
        settings = process_manager.service(ConfigObject)
        ds = process_manager.service(DataStore)

        # used in code testing only
        if settings.get('TESTING'):
            self.logger.warning('Running in TESTING mode. NOT clearing datastore for testing purposes.')
            return StatusCode.Success

        # delete specific items
        for key in self.deletion_keys:
            if key in ds:
                self.logger.debug('Now deleting datastore object with key "{key}".', key=key)
                del ds[key]

        # delete specific class types
        for cls in self.deletion_classes:
            for key in ds:
                if isinstance(ds[key], cls):
                    self.logger.debug('Now deleting datastore object with key "{key}".', key=key)
                    del ds[key]

        # delete all but specific items
        if len(self.keep_only):
            keys = list(ds.keys())
            for key in keys:
                if key not in self.keep_only:
                    self.logger.debug('Now deleting datastore object with key "{key}".', key=key)
                    del ds[key]

        # delete all items in datastore
        if self.clear_all:
            keys = list(ds.keys())
            for key in keys:
                self.logger.debug('Now deleting datastore object with key "{key}".', key=key)
                del ds[key]

        return StatusCode.Success



"""Project: Eskapade - A Python-based package for data analysis.

Module: data_quality.dq_helper

Created: 2017/04/11

Description:
    Data-quality helper functions

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import ast
import re

import numpy as np
import pandas as pd


def cleanup_string(col):
    """Cleanup input string.

    :param col: string to be cleaned up
    :returns: cleaned up string
    :rtype: str
    """
    if not isinstance(col, str):
        return col
    # strip
    new_col = col.strip()
    # replace empty spaces
    new_col = new_col.replace(' ', '_')
    # keep only alphanumeric and _
    new_col = re.sub('[^A-Za-z0-9_]+', '', new_col)
    return new_col


def check_nan(val):
    """Check input value for not a number.

    :param val: value to be checked for nan
    :returns: true if nan
    :rtype: bool
    """
    if pd.isnull(val):
        return True
    if isinstance(val, str):
        val = val.strip()
        if not val or val.lower() == 'none' or val.lower() == 'nan':
            return True
    # from numpy import datetime64
    # if isinstance(val, datetime64):
    #    return val == datetime64('NaT')
    return False


def convert(val):
    """Convert input to interpreted data type.

    :param val: value to be interpreted
    :returns: interpreted value
    """
    try:
        return ast.literal_eval(val)
    except BaseException:
        pass
    return val


def to_str(val, **kwargs):
    """Convert input to string.

    :param val: value to be converted
    :returns: converted value
    :rtype: str
    """
    try:
        if pd.isnull(val):
            return kwargs['nan']
    except BaseException:
        pass
    if isinstance(val, str):
        return val
    if kwargs.get('convert_inconsistent_dtypes', True):
        if hasattr(val, '__str__'):
            return str(val)
    return kwargs['nan']


def to_int(val, **kwargs):
    """Convert input to int.

    :param val: value to be evaluated
    :returns: evaluated value
    :rtype: np.int64
    """
    try:
        if pd.isnull(val):
            return kwargs['nan']
    except BaseException:
        pass
    if isinstance(val, (np.int64, int)):
        return np.int64(val)
    if kwargs.get('convert_inconsistent_dtypes', True):
        try:
            return np.int64(val)
        except BaseException:
            pass
    return kwargs['nan']


def to_float(val, **kwargs):
    """Convert input to float.

    :param val: value to be evaluated
    :returns: evaluated value
    :rtype: np.float64
    """
    try:
        if pd.isnull(val):
            return kwargs['nan']
    except BaseException:
        pass
    if isinstance(val, (np.float64, float)):
        return np.float64(val)
    if kwargs.get('convert_inconsistent_dtypes', True):
        try:
            return np.float64(val)
        except BaseException:
            pass
    return kwargs['nan']


def to_date_time(val, **kwargs):
    """Convert input to numpy.datetime64.

    :param val: value to be evaluated
    :returns: evaluated value
    :rtype: numpy.datetime64
    """
    return pd.to_datetime(val, errors='coerce')


def bool_to_str(val, **kwargs):
    """Convert input boolean to str.

    :param val: value to be evaluated
    :returns: evaluated value
    :rtype: str
    """
    try:
        if pd.isnull(val):
            return kwargs['nan']
    except BaseException:
        pass
    if isinstance(val, (np.bool_, bool)):
        return str(val)
    if kwargs.get('convert_inconsistent_dtypes', True):
        if hasattr(val, '__str__'):
            return str(val)
    return kwargs['nan']


def bool_to_int(val, **kwargs):
    """Convert input boolean to int.

    :param val: value to be evaluated
    :returns: evaluated value
    :rtype: np.int64
    """
    try:
        if pd.isnull(val):
            return kwargs['nan']
    except BaseException:
        pass
    if isinstance(val, (np.bool_, bool)):
        return np.int64(val)
    if kwargs.get('convert_inconsistent_dtypes', False):
        try:
            return np.int64(val)
        except BaseException:
            pass
    return kwargs['nan']


CONV_FUNCS = {str: to_str,
              int: to_int,
              np.int64: to_int,
              bool: bool_to_str,
              np.bool_: bool_to_str,
              float: to_float,
              np.float64: to_float,
              np.datetime64: to_date_time}



"""Project: Eskapade - A python-based package for data analysis.

Macro: Tutorial_1

Created: 2017/02/18

Description:
    Macro illustrates basic setup of chains and links,
    by showing: how to open and run over a dataset,
    apply transformations to it, and plot the results.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import pandas as pd

from eskapade import analysis, process_manager, visualization, ConfigObject, Chain
from eskapade.logger import Logger

logger = Logger()

#########################################################################################

msg = r"""

Be sure to download the input dataset:

$ wget https://s3-eu-west-1.amazonaws.com/kpmg-eskapade-share/data/LAozone.data
"""
logger.info(msg)

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'Tutorial_1'

#########################################################################################
# --- analysis values, settings, helper functions, configuration flags.

VAR_LABELS = dict(doy='Day of year', date='Date', vis='Visibility', vis_km='Visibility')
VAR_UNITS = dict(vis='mi', vis_km='km')


def comp_date(day):
    """Get date/time from day of year."""
    import pandas as pd
    return pd.Timestamp('1976-01-01') + pd.Timedelta('{:d}D'.format(day - 1))


def mi_to_km(dist):
    """Convert miles to kilometres."""
    return dist * 1.60934


conv_funcs = [{'func': comp_date, 'colin': 'doy', 'colout': 'date'},
              {'func': mi_to_km, 'colin': 'vis', 'colout': 'vis_km'},
              # {'func': mph_to_kph, 'colin': , 'colout': 'wind_kph'},
              # {'func': F_to_C, 'colin': , 'colout': 'temp_c'},
              ]

#########################################################################################
# --- now set up the chains and links based on configuration flags

# create first chain
data = Chain('Data')

# add data-frame reader to "Data" chain
reader = analysis.ReadToDf(name='Read_LA_ozone', path='LAozone.data', reader=pd.read_csv, key='data')
data.add(reader)

# add conversion functions to "Data" chain
transform = analysis.ApplyFuncToDf(name='Transform', read_key=reader.key, store_key='transformed_data',
                                   apply_funcs=conv_funcs)
data.add(transform)

# create second chain
summary = Chain('Summary')

# add data-frame summary link to "Summary" chain
summarizer = visualization.DfSummary(name='Create_stats_overview', read_key=transform.store_key,
                                     var_labels=VAR_LABELS, var_units=VAR_UNITS)
summary.add(summarizer)


#########################################################################################
# --- exercises
#
# 1.
# Run the macro and take a look at the output.

# 2.
# Now add your own transformation to the ApplyFuncToDf class.
# We want to transform the temperature to Celsius, so use the code in the comments and fill it out.
# As you can see the output will be written in the DataFrame to the column 'temp_c'.
# We also want to include in the plot the wind speed in km/h.
# Rerun the macro and take a look at the output. The output can be found in decision_engine/results

# 3.
# We are going to make a new link by calling eskapade_generate_link command.
# Place the link in links/tutoriallinks and since the link will be doing a transformation, name it something
# appropriate. Write a link that calls the datastore, picks up the dataframe and adds a new column that contains the
# wind speed in km/h. If this works try to add the temperature in degrees Celsius.

# 4.
# Now we add a new link using templates. In the tutorial you can read how.
# Now we are going to make a completely different transformation in the Link and apply it to the object
# in the DataStore. We want to do add a column to the data that states how humid it is.
# When column 'humidity' is less than 50 it is 'dry', otherwise it is 'humid'.
# You will have to use some pandas functionality or perhaps something else if you prefer. Save the
# new column back into the DataFrame and then put the DataFrame in the DataStore under the key 'data_new'.
# DO NOT USE THE ApplyFuncToDf FOR THIS.

# 5.
# Now rerun the full macro and compare the output with the original output. If you have overwritten over the original
# output, you can add version with settings['version'] = 1 to generate a new output folder.



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk108_reduce

Created: 2017/02/20

Description:
    Macro to illustrate how input lines can be read in,
    processed, and reprinted. E.g. for use in map reduce application.
    Use in combination with: esk108_map

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject
from eskapade import process_manager, resources

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# turning on this flag, the datastore and configuration are not written out to disk
# at the end of the program.

settings = process_manager.service(ConfigObject)
settings['do_reduce'] = True
settings['doNotStoreResults'] = True

#########################################################################################
# --- now parse the follow-up macro

# the flag doNotStoreResults is picked up when parsing the following macro
process_manager.execute_macro(resources.tutorial('esk108_eventlooper.py'))



"""Project: Eskapade - A python-based package for data analysis.

Class: ImportDataStore

Created: 2018-03-17

Description:
    Algorithm to import datastore from external pickle file

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import os
from eskapade import process_manager, DataStore, Link, StatusCode


class ImportDataStore(Link):

    """Link to import datastore from external pickle file.

    Import can happen at initialize() or execute(). Default is initialize() 
    """

    def __init__(self, **kwargs):
        """Initialize an instance of the datastore importer link.

        :param str name: name of link
        :param str path: path of the datastore pickle file to import
        :param bool import_at_initialize: if false, perform datastore import at execute. Default is true, at initialize. 
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'ImportDataStore'))

        # Process and register keyword arguments. If the arguments are not given, all arguments are popped from
        # kwargs and added as attributes of the link. Otherwise, only the provided arguments are processed.
        self._process_kwargs(kwargs, path='', import_at_initialize=True)

        # check residual kwargs; exit if any present
        self.check_extra_kwargs(kwargs)

    def initialize(self):
        """Initialize the link.

        :returns: status code of execution
        :rtype: StatusCode
        """
        # check input arguments
        self.check_arg_types(path=str)
        self.check_arg_vals('path')

        if not os.path.exists(self.path):
            self.logger.fatal('File path "{path}" does not exist.', path=self.path)
            raise AssertionError('Unable to find input file.')

        if self.import_at_initialize:
            self.import_and_update_datastore()
        return StatusCode.Success

    def import_and_update_datastore(self):
        """Import and update the datastore
        """
        # loading external datastore
        ext_store = DataStore.import_from_file(self.path)
        if not isinstance(ext_store, DataStore):
            self.logger.fatal('Object in file "{path}" not of type DataStore.', path=self.path)
            raise AssertionError('Input object not of type DataStore.')
        # overwriting existing datastore
        ds = process_manager.service(DataStore)
        ds.clear()
        ds.update(ext_store)

    def execute(self):
        """Execute the link.

        :returns: status code of execution
        :rtype: StatusCode
        """
        if not self.import_at_initialize:
            self.import_and_update_datastore()        
        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Classes: ArrayStats, GroupByStats

Created: 2017/03/21

Description:
    Summary of an array.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import operator
from collections import Counter

import numpy as np
import pandas as pd
import tabulate
from statsmodels.stats.weightstats import DescrStatsW

from eskapade.analysis.histogram import BinningUtil
from eskapade.logger import Logger

NUM_NS_DAY = 24 * 3600 * int(1e9)


class ArrayStats:
    """Create summary of an array.

    Class to calculate statistics (mean, standard deviation, percentiles,
    etc.) and create a histogram of values in an array.
    The statistics can be returned as values in a dictionary, a
    printable string, or as a LaTeX string.
    """

    logger = Logger()

    def __init__(self, data, col_name, weights=None, unit='', label=''):
        """Initialize for a single column in data frame.

        :param data: Input array
        :type data: iterable
        :type data: pandas Dataframe
        :type data: (keys of) dict
        :param col_name: column name
        :param weights: Input array (default None)
        :type weights: iterable
        :type weights: string (column of data)
        :param unit: Unit of column
        :param str label: Label to describe column variable
        :raises: TypeError
        """
        # set initial values of attributes
        self.stat_vars = []
        self.stat_vals = {}
        self.print_lines = []
        self.latex_table = []

        # parse arguments
        self.name = str(col_name)
        self.unit = str(unit)
        self.label = str(label)
        self.col = data[self.name] if isinstance(data, pd.DataFrame) else data
        self.weights = data[weights] if isinstance(weights, str) else weights
        if isinstance(data, dict):
            self.col = sorted(data.keys())
            self.weights = [data[k] for k in self.col]

        # check if column is iterable
        try:
            iter(self.col)
        except TypeError:
            raise TypeError('Specified data object is not iterable.')
        if self.weights is not None:
            try:
                iter(self.weights)
            except TypeError:
                raise TypeError('Specified weights object is not iterable.')

        # check sizes of data and weights
        if self.weights is not None and len(self.col) != len(self.weights):
            raise AssertionError('weights and data do not have the same length.')

        # store data and weights in a Pandas Series
        if not isinstance(self.col, pd.Series):
            self.col = pd.Series(val for val in self.col)
        if self.weights is not None:
            if not isinstance(self.weights, pd.Series):
                self.weights = pd.Series(w for w in self.weights)

        # store non-null column values
        self.col_nn = self.col[self.col.notnull()]
        self.weights_nn = self.weights[self.col.notnull()] if self.weights is not None else None

        # to be filled in make_histogram
        self.hist = None

    def get_col_props(self):
        """Get column properties.

        :returns dict: Column properties
        """
        return get_col_props(self.col.dtype)

    def create_mpv_stat(self):
        """Compute most probable value from histogram.

        This function computes the most probable value based on the histogram
        from make_histogram(), and adds it to the statistics.
        """
        # basic checks
        if self.hist is None:
            self.logger.warning('Internal histogram is not filled. Run make_histogram() first.')
            return
        if len(self.hist) != 2:
            raise AssertionError('Internal histogram needs to consist of two arrays.')
        values, bins = self.hist
        if not isinstance(values, np.ndarray) and not isinstance(values, list):
            raise TypeError('Values should be a list or numpy array.')
        if not isinstance(bins, np.ndarray) and not isinstance(bins, list):
            raise TypeError('Bins should be a list or numpy array.')
        if len(bins) != len(values) and len(bins) != len(values) + 1:
            raise AssertionError('Bins and values have inconsistent lengths.')

        # if two max elements are equal, this will return the element with the lowest index.
        max_idx = max(enumerate(values), key=lambda x: x[1])[0]
        bc = bins[max_idx]

        # determine column properties
        col_props = self.get_col_props()

        if col_props['is_num'] and len(bins) == len(values) + 1:
            # shift to bin center. note: this also works for timestamps.
            bc += (bins[max_idx + 1] - bc) / 2

        # append statistics
        mpv_name = 'mpv'
        self.stat_vars.append(mpv_name)
        self.stat_vals[mpv_name] = (bc, '{!s}'.format(bc))
        name_len = max(len(n) for n in self.stat_vars)
        self.print_lines.append('{{0:{:d}s}} : {{1:s}}'.format(name_len)
                                .format(mpv_name, self.stat_vals[mpv_name][1]))

    def create_stats(self):
        """Compute statistical properties of column variable.

        This function computes the statistical properties of values in the
        specified column.  It is called by other functions that use the
        resulting figures to create a statistical overview.
        """
        # reset stats containers
        self.stat_vars = []
        self.stat_vals = {}
        self.print_lines = []
        self.latex_table = []

        # determine column properties
        col_props = self.get_col_props()

        # get value counts
        cnt, var_cnt, dist_cnt = (len(self.col), len(self.col_nn), self.col.nunique())
        if self.weights_nn is not None:
            cnt, var_cnt = int(sum(self.weights)), int(sum(self.weights_nn))
        for stat_var, stat_val in zip(('count', 'filled', 'distinct'), (cnt, var_cnt, dist_cnt)):
            self.stat_vars.append(stat_var)
            self.stat_vals[stat_var] = (stat_val, '{:d}'.format(stat_val))
        n_nan = self.col.isnull().sum()
        if n_nan:
            self.stat_vars.append('nan')
            self.stat_vals['nan'] = (n_nan, '{:d}'.format(n_nan))
        # add value counts to print lines
        self.print_lines.append('{}:'.format(self.label if self.label else self.name))
        ratio = (var_cnt / cnt) * 100 if cnt != 0 else 0
        self.print_lines.append('{0:d} entries ({1:.0f}%)'.format(var_cnt, ratio))
        self.print_lines.append('{0:d} unique entries'.format(dist_cnt))

        # convert time stamps to integers
        if col_props['is_ts']:
            col_num = self.col_nn.astype(int)
        else:
            col_num = self.col_nn

        # get additional statistics for numeric variables
        if col_props['is_num'] and len(col_num):
            stat_vars = ('mean', 'std', 'min', 'max', 'p01', 'p05', 'p16', 'p50', 'p84', 'p95', 'p99')
            quant_probs = (0, 1, 0.01, 0.05, 0.16, 0.50, 0.84, 0.95, 0.99)
            # stat_vals = (col_num.mean(), col_num.std(), col_num.min(), col_num.max())\
            #            + tuple(col_num.quantile((0.01, 0.05, 0.16, 0.50, 0.84, 0.95, 0.99)))
            # two lines below also work if weights are None
            des = DescrStatsW(col_num, self.weights_nn)
            stat_vals = (des.mean, des.std) + tuple(weighted_quantile(col_num, self.weights_nn, quant_probs))
            self.stat_vars += stat_vars
            for stat_var, stat_val in zip(stat_vars, stat_vals):
                if not col_props['is_ts']:
                    # value entry for floats and integers
                    self.stat_vals[stat_var] = (stat_val, '{:+g}'.format(stat_val))
                else:
                    if stat_var != 'std':
                        # display time stamps as date/time strings
                        self.stat_vals[stat_var] = (pd.Timestamp(int(stat_val)), str(pd.Timestamp(int(stat_val))))
                    else:
                        # display time-stamp range as number of days
                        stat_val /= NUM_NS_DAY
                        self.stat_vals[stat_var] = (stat_val, '{:g}'.format(stat_val))

            # append statistics to print lines
            name_len = max(len(n) for n in stat_vars)
            for stat_var in stat_vars:
                self.print_lines.append('{{0:{:d}s}} : {{1:s}}'.format(name_len)
                                        .format(stat_var, self.stat_vals[stat_var][1]))

    def get_print_stats(self, to_output=False):
        """Get statistics in printable form.

        :param bool to_output: Print statistics to output stream?
        :returns str: Printable statistics string
        """
        # create statistics print lines
        if not self.stat_vals:
            self.create_stats()

        # create printable string
        print_str = '\n'.join(self.print_lines) + '\n'
        if to_output:
            print(print_str)

        return print_str

    def get_latex_table(self, get_stats=None, latex=True):
        """Get LaTeX code string for table of stats values.

        :param list get_stats: List of statistics that you want to filter on. (default None (all stats))
                               Available stats are: 'count', 'filled', 'distinct', 'mean', 'std', 'min', 'max',
                               'p05', 'p16', 'p50', 'p84', 'p95', 'p99'
        :param bool latex: LaTeX output or list output (default True)
        :returns str: LaTeX code snippet
        """
        # create statistics print lines
        if not self.stat_vals:
            self.create_stats()

        # create LaTeX string
        table = [(stat_var, self.stat_vals[stat_var][1]) for stat_var in self.stat_vars]
        if get_stats is not None:
            table = [(var, val) for var, val in table if var in get_stats]
        if latex:
            return tabulate.tabulate(table, tablefmt='latex')
        else:
            return table

    def get_x_label(self):
        """Get x label."""
        x_lab = self.label if self.label else self.name
        if self.unit:
            x_lab += ' [{}]'.format(self.unit)
        return x_lab

    def make_histogram(self, var_bins=30, var_range=None, bin_edges=None, create_mpv_stat=True):
        """Create histogram of column values.

        :param int var_bins: Number of histogram bins
        :param tuple var_range: Range of histogram variable
        :param list bin_edges: predefined bin edges to use for histogram. Overrules var_bins.
        """
        # create statistics overview
        if not self.stat_vals:
            self.create_stats()

        # determine column properties
        col_props = self.get_col_props()

        if col_props['is_num']:
            col_num = self.col_nn

            # determine histogram range for numeric variable
            if var_range:
                # get minimum and maximum of variable for histogram from specified range
                var_min, var_max = var_range
                if col_props['is_ts']:
                    # convert minimum and maximum to Unix time stamps
                    var_min, var_max = pd.Timestamp(var_min).value, pd.Timestamp(var_max).value
            else:
                # determine minimum and maximum of variable for histogram from percentiles
                var_min, var_max = self.stat_vals.get('p05')[0], self.stat_vals.get('p95')[0]
                if col_props['is_ts']:
                    var_min, var_max = pd.Timestamp(var_min).value, pd.Timestamp(var_max).value
                var_min -= 0.05 * (var_max - var_min)
                var_max += 0.05 * (var_max - var_min)
                if 0. < var_min < +0.2 * (var_max - var_min):
                    var_min = 0.
                elif -0.2 * (var_max - var_min) < var_max < 0.:
                    var_max = 0.
                if np.isnan(var_min):
                    var_min = bin_edges[0]
                assert not np.isnan(var_min)
                if np.isnan(var_max):
                    var_max = bin_edges[-1]
                assert not np.isnan(var_max)

            if col_props['is_ts']:
                # np.histogram cannot deal with timestamps, so convert to ints and convert them back below.
                to_timestamp = np.vectorize(lambda x: pd.Timestamp(x).value)
                col_num = to_timestamp(self.col_nn)
                if bin_edges is not None:
                    bin_edges = (to_timestamp(bin_edges)).tolist()

            if bin_edges is not None:
                bin_util = BinningUtil(bin_edges=bin_edges)
                idx_min = bin_util.value_to_bin_label(var_min)
                var_min = bin_util.get_left_bin_edge(idx_min)
                idx_max = bin_util.value_to_bin_label(var_max)
                var_max = bin_util.get_right_bin_edge(idx_max)
                var_bins = bin_util.truncated_bin_edges(variable_range=[var_min, var_max])
            else:
                if col_props['is_int'] or col_props['is_ts']:
                    # for ints and ts use bins around integer values
                    bin_width = np.max((np.round((var_max - var_min) / float(var_bins)), 1.))
                    var_min = np.floor(var_min - 0.5) + 0.5
                    var_bins = int((var_max - var_min) // bin_width) + int((var_max - var_min) % bin_width > 0.)
                    var_max = var_min + var_bins * bin_width

            # make (weighted) histogram. note that var_bins supersedes range
            values, bins = np.histogram(col_num, bins=var_bins, range=(var_min, var_max), weights=self.weights_nn)

            if col_props['is_ts']:
                # convert Unix time stamps to Pandas time stamps
                bins = [pd.Timestamp(ts) for ts in bins]
            self.hist = values, bins
        else:
            # get data from data frame for categorical column
            if self.weights_nn is None:
                val_counts = self.col_nn.value_counts(sort=True).iloc[:var_bins].to_dict()
            else:
                val_counts = Counter()
                for k, v in zip(self.col_nn, self.weights_nn):
                    val_counts[k] += v
                val_counts = dict(val_counts.most_common(var_bins))
            sorted_vc = sorted(val_counts.items(), key=operator.itemgetter(1), reverse=True)
            labels = [lc[0] for lc in sorted_vc]
            values = [lc[1] for lc in sorted_vc]
            self.hist = values, labels

        # compute most probable value from histogram and add to statistics
        if create_mpv_stat:
            self.create_mpv_stat()

        return self.hist


class GroupByStats(ArrayStats):
    """Create summary of an array in groups."""

    def __init__(self, data, col_name, groupby=None, weights=None, unit='', label=''):
        """Initialize for a single column in dataframe.

        :param data: Input array
        :type data: iterable
        :type data: pandas Dataframe
        :type data: (keys of) dict
        :param col_name: column name
        :param weights: Input array (default None)
        :type weights: iterable
        :type weights: string (column of data)
        :param unit: Unit of column
        :param str label: Label to describe column variable
        :param groupby: column name
        :raises: TypeError
        """
        if groupby:
            assert groupby in data.columns, 'The groupby column is not in your DataFrame'
            group = data.groupby(by=groupby)
            self.stats_obj = {group_key: ArrayStats(group_df, col_name, weights=weights, unit=unit, label=label)
                              for group_key, group_df in group}
        else:
            raise Exception('This class is a wrapper for group-by input. Please supply a proper "groupby" key word.')

    def get_latex_table(self, get_stats=None):
        """Get LaTeX code string for group-by table of stats values.

        :param list get_stats: same as ArrayStats.get_latex_table get_stats key word.
        :returns str: LaTeX code snippet
        """
        # Explicitly strip the tables of their LaTeX headers and footers, concatenate the strings with a group-header
        # and reattach the LaTeX header and footer so that it parses to proper LaTeX.
        self.table = []
        for group_key in sorted(self.stats_obj.keys()):
            s = self.stats_obj.get(group_key).get_latex_table(get_stats=get_stats, latex=False)
            self.table = self.table + [[group_key]] + s  # The double list is needed for proper latex parsing

        if len(self.table) > 25:
            self.logger.warning('The table is longer than 25 rows, the latex file will overflow.')

        return tabulate.tabulate(self.table, tablefmt='latex')


def get_col_props(var_type):
    """Get column properties.

    :returns dict: Column properties
    """
    npdtype = np.dtype(var_type)

    # determine data-type categories
    is_int = isinstance(npdtype.type(), np.integer)
    is_ts = isinstance(npdtype.type(), np.datetime64)
    is_num = is_ts or isinstance(npdtype.type(), np.number)

    return dict(dtype=npdtype, is_num=is_num, is_int=is_int, is_ts=is_ts)


def weighted_quantile(data, weights=None, probability=0.5):
    """Compute the weighted quantile of a 1D numpy array.

    Weighted quantiles, inspired by:
    https://github.com/nudomarinero/wquantiles/blob/master/wquantiles.py
    written by Jose Sabater
    Here updated to return multiple quantiles in one go.  Now also works
    when weight is None.

    :param ndarray data: input array (one dimension).
    :param ndarray weights: array with the weights of the same size of `data`.
    :param ndarray probability: array of quantiles to compute. Each probablity must have a value between 0 and 1.
    :returns: list of the output value(s).
    """
    # Check the inputs
    if not isinstance(data, np.ndarray):
        data = np.asarray(data)
    if isinstance(probability, float):
        probability = [probability]
    if not isinstance(probability, np.ndarray):
        probability = np.asarray(probability)
    for p in probability:
        if p > 1 or p < 0:
            raise ValueError("Probability must have a value between 0 and 1.")
    if data.ndim != 1:
        raise TypeError("Data must be a one dimensional array")
    if weights is None:
        weights = np.ones(len(data))
    if not isinstance(weights, np.ndarray):
        weights = np.asarray(weights)
    if weights.ndim != 1:
        raise TypeError("Weights must be a one dimensional array.")
    if data.shape != weights.shape:
        raise TypeError("The length of data and weights must be the same.")

    # Sort data and compute auxiliary arrays
    sorted_index = np.argsort(data)
    sorted_data = data[sorted_index]
    sorted_weights = weights[sorted_index]
    cumsum = np.cumsum(sorted_weights)
    pn = (cumsum - 0.5 * sorted_weights) / np.sum(sorted_weights)
    # Get the values of the quantiles
    quantiles = [np.interp(p, pn, sorted_data) for p in probability]

    return quantiles



"""Project: Eskapade - A python-based package for data analysis.

Class: ReadToDf

Created: 2016/11/08

Description:
    Algorithm to write pandas dataframes picked up from the datastore.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import copy
import glob
import os

import numpy as np
import pandas as pd

from eskapade import ConfigObject
from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

pd_readers = {'csv': pd.read_csv,
              'tsv': pd.read_csv,
              'xls': pd.read_excel,
              'xlsx': pd.read_excel,
              'json': pd.read_json,
              'h5': pd.read_hdf,
              'sql': pd.read_sql,
              'htm': pd.read_html,
              'html': pd.read_html,
              'dta': pd.read_stata,
              'pkl': pd.read_pickle,
              'pickle': pd.read_pickle}


class ReadToDf(Link):
    """Reads input file(s) to a pandas dataframe.

    You give the link a path where your file is located and some kwargs that go into
    a pandas DataFrame. The kwargs are passed into the file reader.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        Store the configuration of link ReadToDf.

        :param str name: Name given to the link
        :param str path: path of your file to read into pandas DataFrame .
        :param str key: storage key for the DataStore.
        :param reader: pandas reader is determined automatically. But can be set by hand, e.g. csv, xlsx.
        :param bool itr_over_files: Iterate over individual files, default is false.
            If false, are files are collected in one dataframe. NB chunksize takes priority!
        :param int chunksize: Default is none. If positive integer then will always iterate.
            chunksize requires pd.read_csv or pd.read_table.
        :param kwargs: all other key word arguments are passed on to the pandas reader.
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'ReadToDf'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs, path='', key='', reader=None, itr_over_files=False, chunksize=None)

        # pass on remaining kwargs to pandas reader
        self.kwargs = copy.deepcopy(kwargs)

        self._paths = None
        self._path_itr = None
        self._current_path = None
        self._latest_data_length = 0
        self._sum_data_length = 0
        self._iterate = False
        self._reader = None
        self._usecols = self.kwargs.get('usecols', [])

    def set_chunk_size(self, size):
        """Set chunksize setting.

        :param size: chunk size
        """
        self.kwargs['chunksize'] = self.chunksize = size

    def initialize(self):
        """Initialize the link."""
        assert isinstance(self.key, str) and self.key, 'Output key not set.'
        assert isinstance(self._usecols, list), 'Usecols not set correctly.'

        # construct and check list of file paths to read
        read_paths = [p for p in self.path] if not isinstance(self.path, str) else [self.path]
        if not read_paths:
            self.logger.fatal('No file path specified for {cls} instance "{name}".',
                              cls=self.__class__.__name__, name=self.name)
            raise RuntimeError('No file path specified to read dataframe from file.')
        if not all(isinstance(p, str) for p in read_paths):
            self.logger.fatal('Not all paths for {cls} instance "{name}" are strings.',
                              cls=self.__class__.__name__, name=self.name)
            raise TypeError('File paths specified to read dataframe from file must be strings.')

        # construct actual paths
        read_paths = [pe for p in read_paths for pe in glob.glob(p)]
        if not read_paths:
            self.logger.fatal('Specified files not found for {cls} instance "{name}".',
                              cls=self.__class__.__name__, name=self.name)
            raise RuntimeError('specified files not found')
        if not all(os.path.isfile(p) for p in read_paths):
            self.logger.fatal('Not all paths for {cls} instance "{name}" are files.',
                              cls=self.__class__.__name__, name=self.name)
            raise RuntimeError('Paths specified to read dataframe from file must be regular files.')

        # set paths to read
        self._paths = np.array(read_paths)
        self._path_itr = np.nditer(self._paths)

        # now determine if file iterator will be used. Will iterate if:
        # 1. chunksize>0.
        if self.chunksize is not None:
            assert isinstance(self.chunksize,
                              int) and self.chunksize > 0, 'Chunksize needs to be set to positive integer.'
            self._iterate = True
            self.logger.info('chunksize = {size:d}. NB chunksize requires pd.read_csv or pd.read_table.',
                             size=self.chunksize)
            # add back chunksize if it was a kwarg, so it's picked up by pandas.
            self.kwargs['chunksize'] = self.chunksize
            self.logger.info('kwargs passed on to pandas reader are: {kwargs}', kwargs=self.kwargs)
        # 2. more than one file path has been set, and self.itr_over_files==True.
        elif len(self._paths) > 1 and self.itr_over_files is True:
            self._iterate = True
        self.logger.info('File and/or chunksize iterator is active: {is_iterate}.', is_iterate=self._iterate)

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Reads the input file(s) and puts the dataframe in the datastore.
        """
        ds = process_manager.service(DataStore)
        settings = process_manager.service(ConfigObject)

        # 1. handle first the case of no iteration. Concatenate into one dataframe.
        if not self._iterate:
            self.logger.debug('Reading datasets from files [{files}]',
                              files=', '.join('"{}"'.format(p) for p in self._paths))
            df = pd.concat(pandasReader(p, self.reader, **self.kwargs) for p in self._paths)
            numentries = len(df.index)
        # 2. handle case where iteration has been turned on
        else:
            # try picking up new dataset from iterator
            df = next(self)
            while self.latest_data_length() == 0 and not self.is_finished():
                df = next(self)

            # at end of loop
            if self.latest_data_length() == 0:
                assert self.is_finished(), 'Got empty dataset but not at end of iterator.'
                # at end of loop, df == None.
                df = pd.DataFrame(columns=self._usecols)

            # do we have more datasets to go?
            # pass this information to the (possible) repeater at the end of chain
            reqstr = 'chainRepeatRequestBy_' + self.name
            settings[reqstr] = not self.is_finished()

            numentries = self.latest_data_length()
            sumentries = self.sum_data_length()
            self.logger.info('Read next <{n:d}> records; summing up to <{sum_n:d}>.', n=numentries, sum_n=sumentries)
            ds['n_sum_' + self.key] = sumentries

        # store dataframe and number of entries
        ds[self.key] = df
        ds['n_' + self.key] = numentries

        return StatusCode.Success

    def is_finished(self) -> bool:
        """Try to assess if looper is done iterating over files.

        Assess if looper is done or if a next dataset is still coming up.
        """
        finished = self._path_itr.finished
        if isinstance(self.chunksize, int) and self.chunksize > 0:
            finished &= (self._latest_data_length < self.chunksize)
        return finished

    def __next__(self):
        """Pass up the next dataset in the loop.

        Next file is either a entire file or a file chunk.
        Bookkeeping is kept uptodate.
        """
        data = self._next()

        # bookkeeping
        try:
            self._latest_data_length = len(data.index)
        except AttributeError:
            self._latest_data_length = 0
        self._sum_data_length += self._latest_data_length

        return data

    def latest_data_length(self):
        """Return length of current dataset."""
        return self._latest_data_length

    def sum_data_length(self):
        """Return sum length of all datasets processed sofar."""
        return self._sum_data_length

    def _next(self):
        """Pass up the next dataset in the loop.

        This is either a entire file or a file chunk.
        """
        data = None

        # 1. input file has already been set (in previous cycle),
        #    and this is still used for chunking.
        if self._reader is not None and isinstance(self._reader, pd.io.parsers.TextFileReader):
            try:
                data = next(self._reader)
                return data
            except StopIteration:
                # TextFileReader throws stopiterator exception at end
                data = None
            except Exception:
                raise Exception('Unexpected error: cannot process next dataset iteration.')

        # 2. trying next file
        # data is still None, setting up a new reader
        if not self._path_itr.finished:
            path = str(self._path_itr[0])
            self._path_itr.iternext()
            try:
                self._reader = pandasReader(path, self.reader, **self.kwargs)
            except Exception:
                self.logger.fatal('Could not read from new path "{path}".', path=path)
                raise
            self._current_path = path
            self.logger.info('Opened new file "{path}".', path=self._current_path)
        else:
            # no new files left to open
            # (data is still None)
            return data

        # 3. new reader has been set up
        # trying the new reader
        if isinstance(self._reader, pd.core.frame.DataFrame):
            # chunksize not provided, so not chunking.
            data = self._reader
            # resetting the reader for next itr
            self._reader = None
        elif isinstance(self._reader, pd.io.parsers.TextFileReader):
            try:
                data = next(self._reader)
            except StopIteration:
                # TextFileReader throws stopiterator exception at end
                data = None
            except Exception:
                raise Exception('Unexpected error: cannot process next dataset iteration.')

        return data


def pandasReader(path, reader, *args, **kwargs):
    """Pick the correct pandas reader.

    Based on provided reader setting, or based on file extension.
    """
    if not reader:
        reader = pd_readers.get(os.path.splitext(path)[1].strip('.'), None)
    if not reader:
        logger.fatal('No suitable reader found for file "{path}".', path=path)
        raise RuntimeError('unable to find suitable Pandas reader.')
    logger.debug('Using Pandas reader "{reader!s}"', reader=reader)
    # If the reader is input as 'csv' by hand, use the lookup, else use the specified reader (as pd.read_X)
    reader = pd_readers.get(reader) if isinstance(reader, str) else reader
    return reader(path, *args, **kwargs)



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017/02/28

Description:
    Utility functions to collect Eskapade python modules
    e.g. functions to get correct Eskapade file paths and env variables

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import numpy as np
import pandas as pd

from eskapade.logger import Logger

NUM_NS_DAY = 24 * 3600 * int(1e9)

logger = Logger()


def plot_histogram(hist, x_label, y_label=None, is_num=True, is_ts=False, pdf_file_name='', top=20):
    """Create and plot histogram of column values.

    :param hist: input numpy histogram = values, bin_edges
    :param str x_label: Label for histogram x-axis
    :param str y_label: Label for histogram y-axis
    :param bool is_num: True if observable to plot is numeric
    :param bool is_ts: True if observable to plot is a timestamp
    :param str pdf_file_name: if set, will store the plot in a pdf file
    :param int top: only print the top 20 characters of x-labels and y-labels. (default is 20)
    """
    # import matplotlib here to prevent import before setting backend in
    # core.execution.eskapade_run
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_pdf import PdfPages

    plt.figure(figsize=(7, 5))

    try:
        hist_values = hist[0]
        hist_bins = hist[1]
    except BaseException:
        raise ValueError('Cannot extract binning and values from input histogram')

    assert hist_values is not None and len(
        hist_values), 'Histogram bin values have not been set.'
    assert hist_bins is not None and len(
        hist_bins), 'Histogram binning has not been set.'

    # basic attribute check: time stamps treated as numeric.
    if is_ts:
        is_num = True

    # plot numeric and time stamps
    if is_num:
        bin_edges = hist_bins
        bin_values = hist_values
        assert len(bin_edges) == len(bin_values) + 1, \
            'bin edges (+ upper edge) and bin values have inconsistent lengths: {:d} vs {:d}.'\
            .format(len(bin_edges), len(bin_values))

        if is_ts:
            # difference in seconds
            be_tsv = [pd.Timestamp(ts).value for ts in bin_edges]
            width = np.diff(be_tsv)
            # pd.Timestamp(ts).value is in ns
            # maplotlib dates have base of 1 day
            width = width / NUM_NS_DAY
        else:
            width = np.diff(bin_edges)

        # plot histogram
        plt.bar(bin_edges[:-1], bin_values, width=width)

        # set x-axis properties
        plt.xlim(min(bin_edges), max(bin_edges))
        plt.xticks(fontsize=12, rotation=90 if is_ts else 0)
    # plot categories
    else:
        labels = hist_bins
        values = hist_values
        assert len(labels) == len(values), \
            'labels and values have different array lengths: {:d} vs {:d}.'.format(len(labels), len(values))

        # plot histogram
        tick_pos = np.arange(len(labels)) + 0.5
        plt.bar(tick_pos - 0.4, values, width=0.8)

        # set x-axis properties
        def xtick(lab):
            """Get x-tick."""
            lab = str(lab)
            if len(lab) > top:
                lab = lab[:17] + '...'
            return lab

        plt.xlim((0., float(len(labels))))
        plt.xticks(tick_pos, [xtick(lab)
                              for lab in labels], fontsize=12, rotation=90)

    # set common histogram properties
    plt.xlabel(x_label, fontsize=14)
    plt.ylabel(
        str(y_label) if y_label is not None else 'Bin count',
        fontsize=14)
    plt.yticks(fontsize=12)
    plt.grid()

    # store plot
    if pdf_file_name:
        pdf_file = PdfPages(pdf_file_name)
        plt.savefig(pdf_file, format='pdf', bbox_inches='tight', pad_inches=0)
        plt.close()
        pdf_file.close()


def plot_2d_histogram(hist, x_lim, y_lim, title, x_label, y_label, pdf_file_name):
    """Plot 2d histogram with matplotlib.

    :param hist: input numpy histogram = x_bin_edges, y_bin_edges, bin_entries_2dgrid
    :param tuple x_lim: range tuple of x-axis (min,max)
    :param tuple y_lim: range tuple of y-axis (min,max)
    :param str title: title of plot
    :param str x_label: Label for histogram x-axis
    :param str y_label: Label for histogram y-axis
    :param str pdf_file_name: if set, will store the plot in a pdf file
    """
    # import matplotlib here to prevent import before setting backend in
    # core.execution.eskapade_run
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_pdf import PdfPages

    plt.figure(figsize=(7, 5))

    try:
        x_ranges = hist[0]
        y_ranges = hist[1]
        grid = hist[2]
    except BaseException:
        raise ValueError('Cannot extract ranges and grid from input histogram')

    ax = plt.gca()
    ax.pcolormesh(x_ranges, y_ranges, grid)
    ax.set_ylim(y_lim)
    ax.set_xlim(x_lim)
    ax.set_title(title)

    plt.xlabel(x_label, fontsize=14)
    plt.ylabel(y_label, fontsize=14)
    plt.grid()

    if pdf_file_name:
        pdf_file = PdfPages(pdf_file_name)
        plt.savefig(pdf_file, format='pdf', bbox_inches='tight', pad_inches=0)
        plt.close()
        pdf_file.close()


def delete_smallstat(df, group_col, statlim=400):
    """Remove low-statistics groups from dataframe.

    Function to make a new DataFrame that removes all groups of group_col that have less than statlim entries.

    :param df: pandas DataFrame
    :param str group_col: name of the column to group on
    :param int statlim: number of entries a group has to have to be statistically significant
    :returns: smaller DataFrame and the number of removed categories
    :rtype: tuple
    """
    sizes = df.groupby(group_col).size()
    group_sizes = list(zip(df.groupby(group_col).size().index, sizes))

    # Check if there are groups that are too small, if not, return the original DataFrame
    number_toosmall = (sizes < statlim).sum()
    if number_toosmall == 0:
        return df, 0

    # Define the selections for the groups that are too small
    i = 0
    for group, lim in group_sizes:
        if lim < statlim:
            if i == 0:
                selstring = (df[group_col] != group)
            else:
                # Kept this line explicit for clarity
                selstring = selstring & (df[group_col] != group)
            i += 1

    # Return the smaller DataFrame
    df_small = df[selstring]
    return df_small, i


def box_plot(
        df, cause_col, result_col='cost', pdf_file_name='', ylim_quant=0.95, ylim_high=None, ylim_low=0, rot=90,
        statlim=400, label_dict=None, title_add='', top=20):
    """Make box plot.

    Function that plots the boxplot of the column df[result_col] in groups of cause_col. This means that
    the DataFrame is grouped-by on the cause column and then the distribution per group is plotted in a boxplot
    using the standard pandas functionality.
    Boxplots with less than statlim (default=400 ) entries in it are automatically removed.

    :param df: pandas DataFrame
    :param str cause_col: name of the column to group on. This can technically be a number, but that is uncommon.
    :param str result_col: column to do the boxplot on
    :param str pdf_file_name: if set, will store the plot in a pdf file
    :param float ylim_quant: the quantile of the y upper limit
    :param float ylim_high: when defined, this limit is used, when not defined, defaults to None and ylim_high is
           determined by ylim_quant
    :param float ylim_low: matplotlib set_ylim lower bound
    :param int rot: matplotlib rot
    :param int statlim: the number of entries that a group is required to have in order to be plotted
    :param dict label_dict: dictionary with labels for the columns, usage example: label_dict={'col_x': 'Time'}
    :param str title_add: string that is added to the automatic title (the y column name)
    :param int top: only print the top 20 characters of x-labels and y-labels. (default is 20)
    """
    # import matplotlib here to prevent import before setting backend in
    # core.execution.eskapade_run
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_pdf import PdfPages

    # Check the number of categories in the cause_col, if this is too large, only plot the top 20.
    if len(df[cause_col].unique()) > top:
        top_x = df[cause_col].value_counts()[:top].index
        df = df[df[cause_col].isin(top_x)]
        logger.warning('The number of categories of column "{col}" is too large, boxplot is not generated.',
                       col=cause_col)

    # Build a figure
    fig = plt.figure(figsize=(8, 6))
    ax1 = fig.add_subplot(111)

    df_small, _ = delete_smallstat(df, cause_col, statlim=statlim)

    # Make boxplots
    df_small.boxplot(column=result_col, by=cause_col, ax=ax1, fontsize=20, rot=rot, grid=True)

    # If columns do not have a pretty name in label_dict, make the label the column name
    try:
        xlabel = label_dict[cause_col]
    except BaseException:
        xlabel = cause_col
    ax1.set_xlabel(xlabel, fontsize=20)

    try:
        title_label = label_dict[result_col] + title_add
    except BaseException:
        title_label = result_col + title_add
    ax1.set_title(title_label, fontsize=20)

    # Label parameters
    ax1.tick_params(axis='both', which='major', labelsize=20)
    subfig = ax1.get_figure()
    subfig.suptitle('')

    # Calculate quantile for y axis limit
    if ylim_high is None:
        ylim_high = df[result_col].dropna().quantile(q=ylim_quant)

    ax1.set_ylim(ylim_low, ylim_high)

    # Put the number of entries in each group as a number above each boxplot
    num_boxes = len(df_small[cause_col].unique())
    pos = np.arange(num_boxes) + 1
    sizes = list(df_small.groupby(cause_col).size())
    upper_labels = [str(np.round(s, 2)) for s in sizes]
    weights = ['bold', 'semibold']
    for tick in range(num_boxes):
        k = tick % 2
        ax1.text(pos[tick], ylim_high - (ylim_high * 0.05), upper_labels[tick], horizontalalignment='center',
                 size='larger', weight=weights[k])

    # 4. store plot
    if pdf_file_name:
        pdf_file = PdfPages(pdf_file_name)
        plt.savefig(
            pdf_file,
            format='pdf',
            bbox_inches='tight',
            pad_inches=0)
        plt.close()
        pdf_file.close()


def plot_correlation_matrix(matrix_colors, x_labels, y_labels, pdf_file_name='',
                            title='correlation', vmin=-1, vmax=1, color_map='RdYlGn', x_label='', y_label='', top=20,
                            matrix_numbers=None, print_both_numbers=True):
    """Create and plot correlation matrix.

    :param matrix_colors: input correlation matrix
    :param list x_labels: Labels for histogram x-axis bins
    :param list y_labels: Labels for histogram y-axis bins
    :param str pdf_file_name: if set, will store the plot in a pdf file
    :param str title: if set, title of the plot
    :param float vmin: minimum value of color legend (default is -1)
    :param float vmax: maximum value of color legend (default is +1)
    :param str x_label: Label for histogram x-axis
    :param str y_label: Label for histogram y-axis
    :param str color_map: color map passed to matplotlib pcolormesh. (default is 'RdYlGn')
    :param int top: only print the top 20 characters of x-labels and y-labels. (default is 20)
    :param matrix_numbers: input matrix used for plotting numbers. (default it matrix_colors)
    """
    # basic matrix checks
    assert matrix_colors.shape[0] == len(y_labels), 'matrix_colors shape inconsistent with number of y-labels'
    assert matrix_colors.shape[1] == len(x_labels), 'matrix_colors shape inconsistent with number of x-labels'
    if matrix_numbers is None:
        matrix_numbers = matrix_colors
        print_both_numbers = False  # only one set of numbers possible
    else:
        assert matrix_numbers.shape[0] == len(y_labels), 'matrix_numbers shape inconsistent with number of y-labels'
        assert matrix_numbers.shape[1] == len(x_labels), 'matrix_numbers shape inconsistent with number of x-labels'

    # import matplotlib here to prevent import before setting backend in
    # core.execution.eskapade_run
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_pdf import PdfPages
    from matplotlib import colors

    fig, ax = plt.subplots(figsize=(7, 5))
    # cmap = 'RdYlGn' #'YlGn'
    norm = colors.Normalize(vmin=vmin, vmax=vmax)
    img = ax.pcolormesh(matrix_colors, cmap=color_map, edgecolor='w', linewidth=1, norm=norm)

    # set x-axis properties
    def tick(lab):
        """Get tick."""
        if isinstance(lab, (float, int)):
            lab = 'NaN' if np.isnan(lab) else '{0:.1f}'.format(lab)
        lab = str(lab)
        if len(lab) > top:
            lab = lab[:17] + '...'
        return lab

    # reduce default fontsizes in case too many labels?
    nlabs = max(len(y_labels), len(x_labels))
    fontsize_factor = 1
    if nlabs >= 10:
        fontsize_factor = 0.55
    if nlabs >= 20:
        fontsize_factor = 0.25

    # make plot look pretty
    ax.set_title(title, fontsize=14 * fontsize_factor)
    ax.set_yticks(np.arange(len(y_labels)) + 0.5)
    ax.set_xticks(np.arange(len(x_labels)) + 0.5)
    ax.set_yticklabels([tick(lab) for lab in y_labels], rotation='horizontal', fontsize=10 * fontsize_factor)
    ax.set_xticklabels([tick(lab) for lab in x_labels], rotation='vertical', fontsize=10 * fontsize_factor)
    if x_label:
        ax.set_xlabel(x_label, fontsize=12 * fontsize_factor)
    if y_label:
        ax.set_ylabel(y_label, fontsize=12 * fontsize_factor)

    fig.colorbar(img)

    # annotate with correlation values
    numbers_set = [matrix_numbers] if not print_both_numbers else [matrix_numbers, matrix_colors]
    for i, _ in enumerate(x_labels):
        for j, _ in enumerate(y_labels):
            point_color = float(matrix_colors[j][i])
            white_cond = (point_color < 0.7 * vmin) or (point_color >= 0.7 * vmax) or np.isnan(point_color)
            y_offset = 0.5
            for m, matrix in enumerate(numbers_set):
                if print_both_numbers:
                    if m == 0:
                        y_offset = 0.7
                    elif m == 1:
                        y_offset = 0.25
                point = float(matrix[j][i])
                if np.isnan(point):
                    label = 'NaN'
                elif print_both_numbers and m == 0: # number of entries
                    pointi = int(matrix[j][i])
                    label = '{0:d}'.format( pointi )
                else:
                    label = '{0:.2f}'.format(point)
                color = 'w' if white_cond else 'k'
                ax.annotate(label, xy=(i + 0.5, j + y_offset), color=color, horizontalalignment='center',
                            verticalalignment='center', fontsize=10 * fontsize_factor)

    # save plot in file
    if pdf_file_name:
        pdf_file = PdfPages(pdf_file_name)
        plt.savefig(pdf_file, format='pdf', bbox_inches='tight', pad_inches=0)
        plt.close()
        pdf_file.close()



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017/02/27

Description:
    Base classes for the building blocks of an Eskapade analysis run:

        - Link:
        - Chain:

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade.core.definitions import StatusCode
from eskapade.core.meta import Processor, ProcessorSequence
from eskapade.core.mixin import ArgumentsMixin, TimerMixin


class Link(Processor, ArgumentsMixin, TimerMixin):
    """Link base class.

    A link defines the content of an algorithm.  Any actual link is derived
    from this base class.

    A link usually does three things:
    - takes data from the datastore
    - does something to it
    - writes data back

    To take from the data store there is a simple function load()
    To write to the data store there is a simple function store()

    Links are added to a chain as follows:

    >>> from eskapade import process_manager
    >>> from eskapade import analysis
    >>>
    >>> # Create a Chain instance. Note that the chain is automatically registered with process manager.
    >>> io_chain = Chain('IO')
    >>>
    >>> # Add a link to the chain
    >>> reader = analysis.ReadToDf(name='CsvReader', key='foo')
    >>> reader.path = 'foo.csv'
    >>> io_chain.add(reader)
    >>>
    >>> # Run everything.
    >>> process_manager.run()
    """

    def __init__(self, name=None):
        """Initialize link."""
        super().__init__(name)

        # read_key is typically the name of the object picked
        # up from the data store read_key can be a string or
        # list of strings.
        self.read_key = None

        # store_key is typically the name of the object put back
        # in the data tore.
        self.store_key = None

        # required kwargs. filled by _process_kwargs
        self._required_vars = []

        # load() return code.
        self.if_input_missing = StatusCode.Failure
        # store() return code.
        self.if_output_exists = StatusCode.Success

    def _process_kwargs(self, kwargs, **name_val):
        """Process the key word arguments.

        :param kwargs: key word arguments to process
        :param str name_val: name of a required kw arg/setting
        """
        super()._process_kwargs(kwargs, **name_val)
        if name_val:
            self._required_vars += list(name_val.keys())

    def summary(self):
        """Print a summary of the main settings of the link."""
        self.logger.debug('Link: {name}', name=self.name)
        for key in sorted(self._required_vars):
            line = '  attr: {} = '.format(key)
            if not hasattr(self, key):
                raise KeyError('{!s} does not contain item "{!s}".'.format(self, key))
            item = getattr(self, key)
            line += type(item) if not hasattr(item, '__str__') else str(item)
            self.logger.debug(line)

    def __ret_data_list(self, ds, dlist):
        """Check if data exist.

        Internal method used by load method.
        """
        dats = []
        stats = []
        for r in dlist:
            try:
                dats.append(ds[r])
                stats.append(StatusCode.Success.value)
            except KeyError:
                self.logger.warning('Some input data did not exist {r!s} {name}', r=r, name=self.name)
                stats.append(self.if_input_missing.value)
                dats.append(None)

        return StatusCode(max(stats)), dats

    def load(self, ds, read_key=None):
        """Read all data from specified source.

        Read_key can either be:

            * one Link: return statuscode, [data_from_link,...]
            * A list of locations: return statuscode, [data,...]
            * A list of links with only one output location: return statuscode, [data,...]
            * A list of links with multiple output locations: return statuscode, [data,[moredata]...]
            * Any mixture of the above

        Do something logical with a statuscode if this data does not exist
        link.if_input_missing = statuscode

        :return: a tuple statuscode, [data in same order as read_key]
        :rtype: (StatusCode,list)
        """
        if read_key is not None:
            rk = read_key
        elif self.read_key is not None:
            rk = self.read_key
        else:
            self.logger.debug('No read_key defined for {name}.', name=self.name)
            return self.if_input_missing, []
        # Handle the case where this is only one link
        if isinstance(rk, Link):
            if rk.store_key is None:
                self.logger.warning('Link has no store_key {link} {name}.', link=rk.name, name=self.name)
                return self.if_input_missing, [None]
            rk = rk.store_key
        # Always treat it as a list
        if isinstance(rk, str):
            rk = [rk]
        # Iterate over this list
        stats = []
        dats = []
        for r in rk:
            try:
                # Handle links which appear in lists ...
                if isinstance(r, Link):
                    if r.store_key is None:
                        self.logger.warning('Link has no store_key {link} {name}.', link=r.name, name=self.name)
                        stats.append(self.if_input_missing.value)
                        dats.append(None)
                        continue
                    r = r.store_key
                    if not isinstance(r, str):
                        stati, dati = self.__ret_data_list(ds, r)
                        stats.append(stati.value)
                        dats.append(dati)
                        continue
                dats.append(ds[r])
                stats.append(StatusCode.Success.value)
            except KeyError:
                self.logger.warning('Some input data did not exist {link!s} {name}.', link=r, name=self.name)
                stats.append(self.if_input_missing.value)
                dats.append(None)

        return StatusCode(max(stats)), dats

    def __check_store_loc(self, ds, loc):
        """Check if a location exists.

        If yes, return a status code, if no, print an error and return
        link.if_output_exists

        :return: status code if location exists
        :rtype: StatusCode
        """
        if not self.if_output_exists.is_success() and loc in ds:
            self.logger.error('Store key already exists, I am not overwriting {loc} {name}.', loc=loc, name=self.name)
            return self.if_output_exists

        return StatusCode.Success

    def store(self, ds, data, store_key=None, force=False):
        """Store data back to datastore.

        Do something logical with a statuscode if this data already exists
        link.if_output_exists = statuscode uses self.store_key.  If self.store_key is
        a list of locations, I must sent a list of the same length here
        """
        if store_key is not None:
            sk = store_key
        elif self.store_key is not None:
            sk = self.store_key
        else:
            raise AttributeError('store_key has not been set for this link, so I cannot store! {}.'.format(self.name))
        if not isinstance(sk, str):
            if len(data) != len(self.store_key):
                raise ValueError('If you want to store multiple things at once, then the length of the things'
                                 ' you want to store must be the same as the length of self.store_key')
        else:
            data = [data]
            sk = [sk]
        stats = []
        for d, k in zip(data, sk):
            stat = StatusCode.Success if force else self.__check_store_loc(ds, k)
            stats.append(stat.value)
            # If it's not a success then it's not going to be stored
            if not stat.is_success():
                continue
            ds[k] = d
            self.logger.debug('Put object "{key}" in data store with type "{type}".', key=k, type=type(d))

        return StatusCode(max(stats))

    def initialize(self) -> StatusCode:
        """Initialize the Link.

        This method may be overridden by the user.

        :return: Status code.
        :type: StatusCode
        """
        return StatusCode.Success

    def execute(self) -> StatusCode:
        """Execute the Link.

        This method may be overridden by the user.

        :return: Status code.
        :rtype: StatusCode
        """
        return StatusCode.Success

    def finalize(self) -> StatusCode:
        """Finalize the Link.

        This method may be overridden by the user.

        :return: Status code.
        :rtype: StatusCode
        """
        return StatusCode.Success

    def run(self) -> StatusCode:
        """Initialize, execute, and finalize the Link in one go.

        This method is useful for testing purposes, e.g. when developing and testing functionality
        of a link stand-alone and interactively.

        It is not used internally by Eskapade, where the functions are called individually by the chain,
        and all links are initialized together before their common execution,
        and all links in the chain are also finalized together, after their common execution.

        :return: Status code.
        :rtype: StatusCode
        """
        status = StatusCode.Success
        if status.is_success():
            status = self.initialize()
        if status.is_success():
            status = self.execute()
        if status.is_success():
            status = self.finalize()
        return status

class Chain(Processor, ProcessorSequence, TimerMixin):
    """Execution Chain.

    A Chain object contains a collection of links with analysis code. The
    links in a chain are executed in the order in which the links have been
    added to the chain.  Typically a chain contains all links related to one
    topic, for example 'validation of a model', or 'data preparation', or
    'data quality checks'.

    >>> from eskapade import process_manager
    >>> from eskapade import Chain
    >>> from eskapade import analysis
    >>>
    >>> # Create an IO chain. This is automatically registered with the process manager.
    >>> io_chain = Chain('Overview')

    And Links are added to a chain as follows:

    >>> # add a link to the chain
    >>> io_chain.add(analysis.ReadToDf(path='foo.csv', key='foo'))
    >>>
    >>> # Run everything.
    >>> process_manager.run()
    """

    def __init__(self, name, process_manager=None):
        """Initialize chain."""
        super().__init__(name)

        self.prev_chain_name = ''  # type: str
        self.enabled = True  # type: bool

        # We register ourselves with the process manager.
        # If none is specified register with the default
        # process manager.
        if process_manager is None:
            from eskapade import process_manager

        process_manager.add(self)

    def __exec(self, phase_callback) -> StatusCode:
        status = StatusCode.Success

        for _ in self:
            status = phase_callback(_)

            # A link may fail during initialization, execution, and finalization.
            if status == StatusCode.Failure:
                self.logger.fatal('Link "{link!s}" returned "{code!s}" in chain "{chain!s}"!',
                                  link=_, code=status, chain=self)
                break
            # A link may request to skip the rest of the processing during initialization and execution.
            # Why would a link decide to skip the chain it is in during execution?
            # When an essential input collection is empty.
            elif status == StatusCode.SkipChain:
                self.logger.warning('Skipping chain "{chain!s} as requested by link "{link!s}"!',
                                    chain=self, link=_)
                break
            # A link may request to skip the rest of the execution of the chain (but do perform finalize).
            elif status == StatusCode.BreakChain:
                self.logger.warning('Breaking of exection of chain "{chain!s} as requested by link "{link!s}"!',
                                    chain=self, link=_)
                break
            # A link may request that the chain needs to be repeated during execution, e.g.
            # when looping over input file in chunks.
            elif status == StatusCode.RepeatChain:
                self.logger.warning('Repeating chain "{chain!s}" as requested by link "{link!s}"!',
                                    chain=self, link=_)
                break
            # Default, is to log an unhandled status code from the chain.
            elif status != StatusCode.Success:
                self.logger.fatal('Unhandled StatusCode "{status!s}" from link "{link!s}" in "{chain!s}"!',
                                  status=status, link=_, chain=self)
                break

        return status

    def add(self, link: Link) -> None:
        """Add a link to the chain.

        :param link: The link to add to the chain.
        :type link: Link
        :raise TypeError: When the link is of an incompatible type.
        :raise KeyError: When a Link of the same type and name already exists.
        """
        if not issubclass(type(link), Link):
            raise TypeError('Expected (sub-class of) "Link" and not "{wrong!s}"!'.format(wrong=type(link)))

        self.logger.debug('Registering link "{link}."', link=link)

        link.parent = self
        super().add(link)

    # TODO (janos4276) This is a copy past of the Process Manager method. We could/should probably move this to
    # ProcesserSequence.
    def get(self, link_name: str) -> Link:
        """Find the link with the given name.

        :param link_name: Find a link with the given name.
        :type link_name: str
        :return: The chain.
        :rtype: Chain
        :raise ValueError: When the given chain name cannot be found.
        """
        iterator = iter(self)
        link = next(iterator, None)
        while link and link.name != link_name:
            link = next(iterator, None)

        if link is None:
            raise ValueError('Found no chain with name "{name}"!'.format(name=link_name))

        return link

    def discard(self, link: Link) -> None:
        """Remove a link from the chain.

        :param link:
        :type link: Link
        :raise KeyError: When the processor does not exist.
        """
        if not issubclass(type(link), Link):
            raise TypeError('Expected (sub-class of) "Link" not "{wrong!s}"!'.format(wrong=type(link)))

        super().discard(link)
        link.parent = None

    def initialize(self) -> StatusCode:
        """Initialize chain and links.

        :return: Initialization status code.
        :rtype: StatusCode
        """
        self.logger.debug('Initializing chain "{chain!s}".', chain=self)

        self.start_timer()

        status = self.__exec(Processor._initialize)

        if status == StatusCode.Success:
            self.logger.debug('Successfully initialized chain "{chain!s}".', chain=self)

        return status

    def execute(self) -> StatusCode:
        """Execute links in chain.

        :return: Execution status code.
        :rtype: StatusCode
        """
        self.logger.debug('Executing chain "{chain!s}".', chain=self)

        status = self.__exec(Processor._execute)

        if status == StatusCode.Success:
            self.logger.debug('Successfully executed chain "{chain!s}".', chain=self)

        return status

    def finalize(self) -> StatusCode:
        """Finalize links and chain.

        :return: Finalization status code.
        :rtype StatusCode:
        """
        self.logger.debug('Finalizing chain "{chain!s}".', chain=self)

        status = self.__exec(Processor._finalize)

        if status == StatusCode.Success:
            total_time = self.stop_timer()
            self.logger.debug('Successfully finalized chain "{chain!s}".', chain=self)
            self.logger.debug('Chain "{chain!s}" took {sec:.2f} seconds to complete.',
                              chain=self, sec=total_time)

        return status

    def clear(self):
        """Clear the chain."""
        self.parent = None
        super().clear()

    @property
    def n_links(self) -> int:
        """Return the number of links in the chain.

        :return: The number of links in the chain.
        :rtype: int
        """
        return len(self)



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk108_eventlooper

Created: 2017/02/20

Description:
    Macro to illustrate how input lines can be read in,
    processed, and reprinted. E.g. for use in map reduce application.
    Used as input for: esk108_map and esk108_reduce

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import tempfile

from eskapade import ConfigObject, Chain
from eskapade import core_ops
from eskapade import process_manager

#########################################################################################
# --- minimal analysis information
settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk108_eventlooper'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

settings['do_map'] = settings.get('do_map', False)
settings['do_reduce'] = settings.get('do_reduce', False)

settings['TESTING'] = settings.get('TESTING', False)

# --- create dummy example dataset, which is used below
if settings['TESTING']:
    tmp = b"""# dataset from:
# https://rajmak.wordpress.com/2013/04/27/clustering-text-map-reduce-in-python/
Converse All Star PC2 - Boys' Toddler
HI Nike Sport Girls Golf Dress
Brooks Nightlife Infiniti 1/2 Zip - Women's
HI Nike Solid Girls Golf Shorts
Nike Therma-FIT K.O. (MLB Rays)
adidas adiPURE IV TRX FG - Men's
Nike College All-Purpose Seasonal Graphic (Oklahoma) Womens T-Shirt
adidas Adipure 11PRO TRX FG - Women's
HI Nike Team (NFL Giants) BCA Womens T-Shirt
adidas Sprintstar 4 - Men's
HI Nike Attitude (NFL Titans) BCA Womens T-Shirt
HI Nike Polo Girls Golf Dress
Nike Therma-FIT K.O. (MLB Twins)
adidas Sprintstar 3 - Women's
Under Armour Performance Team Polo - Mens - For All Sports - Clothing - Purple/White
Converse All Star Ox - Girls' Toddler
HI Nike College All-Purpose Seasonal Graphic (Washington) Womens T-Shirt
Under Armour Performance Team Polo - Mens - For All Sports - Clothing - Red/White
Nike Therma-FIT K.O. (MLB Phillies)
Brooks Nightlife Infiniti 1/2 Zip Jacket - Mens
"""
    f = tempfile.NamedTemporaryFile(delete=False)
    f.write(tmp)
    f.close()
    # file is not immediately deleted because we used delete=False
    # used below with f.name


def to_lower(x):
    """Lower case."""
    return x.lower()


def first_word(x):
    """Take first word."""
    return x.split()[0]


#########################################################################################
# --- now set up the chains and links based on configuration flags

# This chain does 'mapping'. (macro B does 'reduction'.)

# --- mapper: chain with event looper
#     this eventlooper link serves as a mapper.
#     in this example the lines are converted to lower chars, and the first word is selected.
if settings['do_map']:
    ch = Chain("Mapper")
    looper = core_ops.EventLooper(name='listener')
    looper.skip_line_beginning_with = ['#']
    looper.line_processor_set = [first_word, to_lower]
    if settings['TESTING']:
        looper.filename = f.name
    ch.add(looper)

# --- reducer: chain with event looper
#     this eventlooper link serves as a reducer
#     in this example the lines are grouped together into unique sets.
if settings['do_reduce']:
    ch = Chain("Reducer")
    looper = core_ops.EventLooper(name='grouper')
    # reducer selects all unique lines
    looper.sort = True
    looper.unique = True
    looper.store_key = 'products'
    if settings['TESTING']:
        looper.filename = f.name
    ch.add(looper)

    # ... do other operations to lines here

    # print lines
    link = core_ops.LinePrinter()
    link.read_key = looper.store_key
    ch.add(link)



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk103_printdatastore

Created: 2017/02/15

Description:
    Macro to illustrate the use of the Printdatastore link.
    Prindatastore prints an overview of the contents in the
    datastore at the state of running

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import core_ops, process_manager, ConfigObject, DataStore, Chain
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk103_printdatastore')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk103_printdatastore'
settings['version'] = 0

#########################################################################################
# --- for this macro, fill the datastore with some dummy information

process_manager.service(DataStore)['hello'] = 'world'
process_manager.service(DataStore)['d'] = {'a': 1, 'b': 2, 'c': 3}

#########################################################################################
# --- now set up the chains and links based on configuration flags

ch = Chain('Overview')

# printdatastore prints an overview of the contents in the datastore
# at the state of executing the link.
# The overview consists of list of keys in the datastore and and the object types.
link = core_ops.PrintDs()
# keys are the items for which the contents of the actual item is printed.
# if the key is not known ('foo'), then it is skipped.
link.keys = ['foo', 'hello', 'd']
ch.add(link)

#########################################################################################

logger.debug('Done parsing configuration file esk103_printdatastore')



"""Project: Eskapade - A python-based package for data analysis.

Class : DfBoxplot

Created: 2017/02/17

Description:
    Link to create a boxplot of data frame columns.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import pandas as pd

from eskapade import StatusCode, DataStore, Link, process_manager
from eskapade import resources, visualization
from eskapade.analysis import statistics
from eskapade.core import persistence

NUMBER_OF_RECORDS = 1000


class DfBoxplot(Link):
    """Create a boxplot of one column of a DataFrame that is grouped by values from a second column.

    Creates a report page for each variable in DataFrame, containing:

    * a profile of the column dataset
    * a nicely scaled plot of the boxplots per group of the column

    Example is available in: tutorials/esk304_df_boxplot.py
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param str read_key: key of input data to read from data store
        :param str results_path: output path of summary result files
        :param str column: column pick up from input data to use as boxplot input
        :param list cause_columns: list of columns (str) to group-by, and per unique value plot a boxplot
        :param list statistics: a list of strings of the statistics you want to generate for the boxplot
               the full list is taken from statistics.ArrayStats.get_latex_table
               defaults to: ['count', 'mean', 'min', 'max']
        :param str pages_key: data store key of existing report pages
        """
        # initialize Link
        Link.__init__(self, kwargs.pop('name', 'df_boxplot'))

        # process keyword arguments
        self._process_kwargs(kwargs, read_key='', results_path='', column=None, cause_columns=None,
                             var_labels={}, var_units={}, statistics=['count', 'mean', 'min', 'max'], pages_key='')
        self.check_extra_kwargs(kwargs)

        # initialize attributes
        self.pages = []

    def _process_results_path(self):
        """Process results_path argument."""
        if not self.results_path:
            self.results_path = persistence.io_path('results_data', 'report')
        persistence.create_dir(self.results_path)

    def initialize(self):
        """Initialize the link."""
        # check input arguments
        self.check_arg_types(read_key=str, pages_key=str)
        self.check_arg_types(recurse=True, allow_none=True, column=str, cause_columns=list, statistics=list)
        self.check_arg_vals('read_key')

        # read report templates, we use the summary_report template from the df summary link
        with open(resources.template('df_summary_report.tex')) as templ_file:
            self.report_template = templ_file.read()
        with open(resources.template('df_summary_report_page.tex')) as templ_file:
            self.page_template = templ_file.read()

        self._process_results_path()

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Creates a report page for each column that we group-by in the data frame.

        * create statistics object for group
        * create overview table of column variable
        * plot boxplot of column variable per group
        * store plot
        """
        # fetch and check input data frame
        ds = process_manager.service(DataStore)
        data = ds.get(self.read_key, None)
        if not isinstance(data, pd.DataFrame):
            self.logger.fatal('No Pandas data frame "{self.read_key}" found in data store for {self!s}', self=self)
            raise RuntimeError('No input data found for {!s}.'.format(self))

        # fetch any existing report pages
        if self.pages_key:
            self.pages = ds.get(self.pages_key, [])
            assert isinstance(self.pages, list), 'Pages key {} does not refer to a list.'.format(self.pages_key)

        # create report page for each plot
        for col in self.cause_columns:
            # output column name
            self.logger.debug('Processing cause column "{col}".', col=col)

            # check if column is in data frame
            if col not in data.columns:
                self.logger.warning('Column "{col}" not in data frame.', col=col)
                continue

            # 1. create statistics object for column
            var_label = self.column

            # 2. Calculate the statistical properties per group
            # Notice that in this link we call GroupByStats and in df_summary we call ArrayStats
            stats = statistics.GroupByStats(data, self.column, groupby=col, unit=self.var_units.get(self.column, ''),
                                            label=var_label)

            # 3. plot and store histogram of column variable
            box_file_name = 'boxplot_{}.pdf'.format(col)
            pdf_file_name = '{0:s}/{1:s}'.format(self.results_path, box_file_name)
            visualization.vis_utils.box_plot(data, col, self.column, pdf_file_name=pdf_file_name)

            # 4. create overview table of column variable with a group-by applied by GroupByStats
            stats_table = stats.get_latex_table(get_stats=self.statistics)

            # 5. create page string
            self.pages.append(self.page_template.replace('VAR_LABEL', var_label)
                              .replace('VAR_STATS_TABLE', stats_table)
                              .replace('VAR_HISTOGRAM_PATH', box_file_name))

        # storage
        if self.pages_key:
            ds[self.pages_key] = self.pages

        return StatusCode.Success

    def finalize(self):
        """Finalize the link."""
        # write report file from the strings in self.pages
        with open('{}/report.tex'.format(self.results_path), 'w') as report_file:
            report_file.write(self.report_template.replace('INPUT_PAGES', ''.join(self.pages)))

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk203_apply_func_to_pandas_df

Created: 2017/02/23

Description:
    Illustrates link that calls basic apply() to columns of a pandas dataframes
    See for more information pandas documentation:

    http://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.apply.html

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""
from numpy.random import randn
from pandas import DataFrame

from eskapade import analysis, core_ops, process_manager, ConfigObject, DataStore, Chain
from eskapade.logger import Logger, LogLevel

logger = Logger()

logger.debug('Now parsing configuration file esk203_apply_func_to_pandas_df')

#########################################################################################
# --- minimal analysis information
settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk203_apply_func_to_pandas_df'
settings['version'] = 0


#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# functions to be applied below

def square(x):
    """Square value."""
    return x * x


def sqrt_abs(x):
    """Take square root of absolute value."""
    from math import sqrt
    return sqrt(abs(x))


conv_funcs = [{'func': square, 'colin': 'x', 'colout': 'xx'},
              {'func': sqrt_abs, 'colin': 'y', 'colout': 'yy'}]

# generate a dummy dataframe and add to datastore
# to this dataset selections are applied below, during link execution.

# NB: realize that, normally, such a dataframe is read or constructed on the fly
# during link execution.

df = DataFrame(randn(20, 2), columns=list('xy'))

ds = process_manager.service(DataStore)
ds['incoming_data'] = df

#########################################################################################
# --- now set up the chains and links based on configuration flags

data_prep = Chain('DataPrep')

# query_set = seletions that are applies to incoming_records
# after selections, only keep column in select_columns ('a', 'c')
# add conversion functions to "Data" chain
link = analysis.ApplyFuncToDf(name='Transform',
                              read_key='incoming_data',
                              store_key='transformed_data',
                              apply_funcs=conv_funcs)
# Any other kwargs given to ApplyFuncToDf are passed on the the
# pandas query() function.
link.logger.log_level = LogLevel.DEBUG
data_prep.add(link)

link = core_ops.DsObjectDeleter()
link.deletion_keys = ['incoming_data']
data_prep.add(link)

link = core_ops.PrintDs()
link.keys = ['transformed_data']
data_prep.add(link)

#########################################################################################

logger.debug('Done parsing configuration file esk203_apply_func_to_pandas_df')



"""Project: Eskapade - A python-based package for data analysis.

Class: SkipChainIfEmpty

Created: 2016/11/08

Description:
    Algorithm to skip to the next Chain if input dataset is empty

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import process_manager, DataStore, Link, StatusCode


class SkipChainIfEmpty(Link):
    """Sends a SkipChain enums.StatusCode signal when an appointed dataset is empty.

    This signal causes that the Processs Manager to step immediately to the next Chain.
    Input collections can be either mongo collections or dataframes in the datastore.
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        :param str name: name of link
        :param list collection_set: datastore keys holding the datasets to be checked. If any of these is empty,
                                    the chain is skipped.
        :param bool skip_chain_when_key_not_in_ds: skip the chain as well if the dataframe is not present in the
                                                   datastore. When True and if type is 'pandas.DataFrame', sends
                                                   a SkipChain signal if key not in DataStore
        :param bool check_at_initialize: perform dataset empty is check at initialize. Default is true.
        :param bool check_at_execute: perform dataset empty is check at initialize. Default is false.
        """
        Link.__init__(self, kwargs.pop('name', 'SkipChainIfEmpty'))

        # process keyword arguments
        self._process_kwargs(kwargs, collection_set=[], skip_chain_when_key_not_in_ds=False, check_at_initialize=True,
                             check_at_execute=False)
        self.check_extra_kwargs(kwargs)

    def initialize(self):
        """Initialize the link."""
        if self.check_at_initialize:
            return self.check_collection_set()

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        Skip to the next Chain if any of the input dataset is empty.
        """
        if self.check_at_execute:
            return self.check_collection_set()

        return StatusCode.Success

    def check_collection_set(self):
        """Check existence of collection in either mongo or datastore, and check that they are not empty.

        Collections need to be both present and not empty.

        - For mongo collections a dedicated filter can be applied before doing the count.
        - For pandas dataframes the additional option 'skip_chain_when_key_not_in_ds' exists. Meaning,
          skip the chain as well if the dataframe is not present in the datastore.
        """
        # check if collection names are present in datastore
        ds = process_manager.service(DataStore)
        for k in self.collection_set:
            if k not in ds:
                if self.skip_chain_when_key_not_in_ds:
                    self.logger.warning('Key {key!s} not in DataStore. Sending skip chain signal.', key=k)
                    return StatusCode.SkipChain
                else:
                    raise Exception('Key "{key}" not in DataStore.'.format(key=k))
            df = ds[k]
            if len(df.index) == 0:
                self.logger.warning(
                    'pandas.DataFrame with datastore key "{key!s}" is empty. Sending skip chain signal.', key=k)
                return StatusCode.SkipChain

        return StatusCode.Success



from eskapade.analysis.links.apply_func_to_df import ApplyFuncToDf
from eskapade.analysis.links.apply_selection_to_df import ApplySelectionToDf
from eskapade.analysis.links.basic_generator import BasicGenerator
from eskapade.analysis.links.df_concatenator import DfConcatenator
from eskapade.analysis.links.df_merger import DfMerger
from eskapade.analysis.links.histogrammar_filler import HistogrammarFiller
from eskapade.analysis.links.random_sample_splitter import RandomSampleSplitter
from eskapade.analysis.links.read_to_df import ReadToDf
from eskapade.analysis.links.record_factorizer import RecordFactorizer
from eskapade.analysis.links.record_vectorizer import RecordVectorizer
from eskapade.analysis.links.value_counter import ValueCounter
from eskapade.analysis.links.write_from_df import WriteFromDf

__all__ = ['ApplyFuncToDf',
           'ApplySelectionToDf',
           'BasicGenerator',
           'DfConcatenator',
           'DfMerger',
           'HistogrammarFiller',
           'RandomSampleSplitter',
           'ReadToDf',
           'RecordFactorizer',
           'RecordVectorizer',
           'ValueCounter',
           'WriteFromDf']



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk106_cmdline_options

Created: 2017/02/20

Description:
    Macro to illustrate the use of flags set from the command line.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain
from eskapade import core_ops
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk106_cmdline_options')

#########################################################################################
# --- minimal analysis information
settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk106_cmdline_options'
settings['version'] = 0

msg = r"""

The two flags below control whether chains are turned on or off. (default=on)
from the cmd line, control these with:

-c do_chain0=False -c do_chain1=False

Try it; No Hello Worlds will be printed.
"""
logger.info(msg)

#########################################################################################
# --- now set up the chains and links based on configuration flags

if settings.get('do_chain0', True):
    ch = Chain('Chain0')
    link = core_ops.HelloWorld(name='hello0')
    link.hello = 'Town'
    ch.add(link)

if settings.get('do_chain1', True):
    ch = Chain('Chain1')
    link = core_ops.HelloWorld(name='hello1')
    link.hello = 'Universe'
    ch.add(link)

#########################################################################################

logger.debug('Done parsing configuration file esk106_cmdline_options')



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk104_datastore_operations

Created: 2017/02/20

Description:
    Macro to illustrate how to control the contents of the datastore

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject, Chain
from eskapade import core_ops
from eskapade import process_manager
from eskapade.logger import Logger

logger = Logger()

logger.debug('Now parsing configuration file esk104_basic_datastore_operations.')

#########################################################################################
# --- minimal analysis information

settings = process_manager.service(ConfigObject)
settings['analysisName'] = 'esk104_basic_datastore_operations'
settings['version'] = 0

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# some dummy information to use in this macro
f = {'hello': 'world', 'v': [3, 1, 4, 1, 5], 'n_favorite': 7}
g = {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 'favorite'}

#########################################################################################
# --- now set up the chains and links based on configuration flags

# demo consists of 5 simple parts:
#
# 1. putting items in the datastore, and displaying the contents.
# 2. asserting the presence of items in the datastore, and then deleting individual items from the datastore.
# 3. deleting all items from the datastore.
# 4. deleting all but certain items from the datastore.
# 5. moving, copying, or removing objects from the datastore

#########
# chain 1:
# - putting items in the datastore.
# - displaying the contents of items in the datastore.

ch = Chain('chain1')

# the link ToDsDict adds objects to the datastore
# by default this happens at the execution of the link.
# (optionally, this can be done at initialization.)
# Here it is used as a dummy data generator.

link = core_ops.ToDsDict(name='intods_1')
link.obj = f
# copydict = true: all items in dict f are added to the datastore
link.copydict = True
ch.add(link)

# print contents of datastore
link = core_ops.PrintDs()
link.keys = ['n_favorite', 'hello']
ch.add(link)

#########
# chain 2
# - asserting the presence of items in the datastore.
# - deleting individual items from the datastore.

ch = Chain('chain2')

# the link AssertInDs checks the presence
# of certain objects in the datastore
link = core_ops.AssertInDs()
link.keySet = ['hello', 'n_favorite']
ch.add(link)

# the link DsObjectDeleter removes objects from the datastore
# here, keep only 'v'
link = core_ops.DsObjectDeleter()
link.keep_only = ['v']
ch.add(link)

link = core_ops.PrintDs()
link.keys = ['v', 'hello']
ch.add(link)

#########
# chain 3
# - deleting all items from the datastore.

ch = Chain('chain3')

# default, delete everything from the datastore
link = core_ops.DsObjectDeleter()
ch.add(link)

link = core_ops.ToDsDict(name='intods_3')
link.obj = g
link.copydict = True
ch.add(link)

link = core_ops.PrintDs()
ch.add(link)

#########
# chain 4
# - deleting all but certain items from the datastore.

ch = Chain('chain4')

link = core_ops.AssertInDs()
link.keySet = ['a', 'b']
ch.add(link)

# here, delete only c and d
link = core_ops.DsObjectDeleter()
link.deletion_keys = ['d', 'e']
ch.add(link)

link = core_ops.PrintDs()
ch.add(link)

#########
# chain 5
# - moving, copying, or removing objects from the datastore

ch = Chain('chain5')

link = core_ops.AssertInDs()
link.keySet = ['a', 'b', 'c']
ch.add(link)

# the link DsToDs can move, copy, or remove objects from the datastore
# default is move.

# in this example, move key 'c' to 'd'
link = core_ops.DsToDs(name='ds_move')
link.read_key = 'c'
link.store_key = 'd'
ch.add(link)

link = core_ops.PrintDs(name='pds1')
link.keys = ['d']
ch.add(link)

# in this example, copy key 'd' to 'e'
link = core_ops.DsToDs(name='ds_copy')
link.read_key = 'd'
link.store_key = 'e'
link.copy = True
ch.add(link)

link = core_ops.PrintDs(name='pds2')
link.keys = ['d', 'e']
ch.add(link)

# in this example, remove item 'd'
link = core_ops.DsToDs(name='ds_remove')
link.read_key = 'd'
link.remove = True
ch.add(link)

link = core_ops.PrintDs(name='pds3')
link.keys = ['e']
ch.add(link)

link = core_ops.DsObjectDeleter()
link.keep_only = ['a']
ch.add(link)

# empty ...
link = core_ops.PrintDs()
link.keys = ['a']
ch.add(link)

#########################################################################################

logger.debug('Done parsing configuration file esk104_basic_datastore_operations.')



"""Project: Eskapade - A python-based package for data analysis.

Created: 2017/08/08

Description:
    setup script.

Authors:
    KPMG Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import logging

from setuptools import find_packages
from setuptools import setup

NAME = 'Eskapade'

MAJOR = 0
REVISION = 8
PATCH = 0
DEV = False

VERSION = '{major}.{revision}.{patch}'.format(major=MAJOR, revision=REVISION, patch=PATCH)
FULL_VERSION = VERSION
if DEV:
    FULL_VERSION += '.dev'

TEST_REQUIREMENTS = ['pytest>=3.5.0',
                     'pytest-pylint>=0.9.0',
                     ]

REQUIREMENTS = [
    'pendulum==1.2.5',
    'matplotlib>=2.0.2',
    'numpy>=1.14.2',
    'scipy>=0.19.0',
    'scikit-learn>=0.18.1',
    'statsmodels>=0.8.0',
    'pandas>=0.20.1',
    'tabulate>=0.8.2',
    'sortedcontainers>=1.5.7',
    'histogrammar>=1.0.9',
    'names>=0.3.0',
    'fastnumbers>=2.0.2'
    ]

REQUIREMENTS = REQUIREMENTS + TEST_REQUIREMENTS

CMD_CLASS = dict()
COMMAND_OPTIONS = dict()

EXCLUDE_PACKAGES = []
EXTERNAL_MODULES = []

logging.basicConfig()
logger = logging.getLogger(__file__)


def write_version_py(filename: str = 'python/eskapade/version.py') -> None:
    """Write package version to version.py.

    This will ensure that the version in version.py is in sync with us.

    :param filename: The version.py to write too.
    :type filename: str
    :return:
    :rtype: None
    """
    # Do not modify the indentation of version_str!
    version_str = """\"\"\"THIS FILE IS AUTO-GENERATED BY ESKAPADE SETUP.PY.\"\"\"

name = '{name!s}'
version = '{version!s}'
full_version = '{full_version!s}'
release = {is_release!s}
"""

    version_file = open(filename, 'w')
    try:
        version_file.write(version_str.format(name=NAME.lower(),
                                              version=VERSION,
                                              full_version=FULL_VERSION,
                                              is_release=not DEV))
    finally:
        version_file.close()


def setup_package() -> None:
    """The main setup method.

    It is responsible for setting up and installing the package.

    :return:
    :rtype: None
    """
    write_version_py()

    setup(name=NAME,
          version=FULL_VERSION,
          url='http://eskapade.kave.io',
          license='',
          author='KPMG N.V. The Netherlands',
          author_email='kave@kpmg.com',
          description='Eskapade modular analytics',
          python_requires='>=3.5',
          package_dir={'': 'python'},
          packages=find_packages(where='python', exclude=EXCLUDE_PACKAGES),
          # Setuptools requires that package data are located inside the package.
          # This is a feature and not a bug, see
          # http://setuptools.readthedocs.io/en/latest/setuptools.html#non-package-data-files
          package_data={
              NAME.lower(): ['config/*', 'templates/*', 'data/*', 'tutorials/*.sh']
          },
          install_requires=REQUIREMENTS,
          tests_require=TEST_REQUIREMENTS,
          ext_modules=EXTERNAL_MODULES,
          cmdclass=CMD_CLASS,
          command_options=COMMAND_OPTIONS,
          # The following 'creates' executable scripts for *nix and Windows.
          # As an added bonus the Windows scripts will auto-magically
          # get a .exe extension.
          #
          # eskapade_run: main application to let loose on macros.
          # eskapade_trial: test application to let loose on tests. This is just a wrapper around pytest.
          # eskapade_generate_link: utility to generate link boilerplate/template.
          # eskapade_generate_macro: utility to generate macro boilerplate/template.
          # eskapade_generate_notebook: utility to generate notebook boilerplate/template.
          # eskapade_bootstrap: utility to bootstrap an Eskapade project.
          entry_points={
              'console_scripts': [
                  'eskapade_ignite = eskapade.entry_points:eskapade_ignite',
                  'eskapade_run = eskapade.entry_points:eskapade_run',
                  'eskapade_trial = eskapade.entry_points:eskapade_trial',
                  'eskapade_generate_link = eskapade.entry_points:eskapade_generate_link',
                  'eskapade_generate_macro = eskapade.entry_points:eskapade_generate_macro',
                  'eskapade_generate_notebook = eskapade.entry_points:eskapade_generate_notebook',
                  'eskapade_bootstrap = eskapade.entry_points:eskapade_bootstrap'
              ]
          }
          )


if __name__ == '__main__':
    setup_package()



"""Project: Eskapade - A python-based package for data analysis.

Class: HistogrammarFiller

Created: 2017/03/21

Description:
    Algorithm to fill histogrammar sparse-bin histograms.
    It is possible to do cleaning of these histograms by
    rejecting certain keys or removing inconsistent data types.
    Timestamp columns are converted to nanoseconds before
    the binning is applied.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import histogrammar as hg
import numpy as np

from eskapade.analysis import histogram_filling as hf
from eskapade.analysis.histogram_filling import HistogramFillerBase


class HistogrammarFiller(HistogramFillerBase):
    """Fill histogrammar sparse-bin histograms.

    Algorithm to fill histogrammar style sparse-bin and category histograms.

    It is possible to do after-filling cleaning of these histograms by rejecting certain
    keys or removing inconsistent data types. Timestamp columns are
    converted to nanoseconds before the binning is applied. Final histograms
    are stored in the datastore.

    Example is available in: tutorials/esk303_hgr_filler_plotter.py
    """

    def __init__(self, **kwargs):
        """Initialize link instance.

        Store and do basic check on the attributes of link HistogrammarFiller.

        :param str name: name of link
        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store histograms in data store
        :param list columns: colums to pick up from input data. (default is all columns)
        :param dict bin_specs: dictionaries used for rebinning numeric or timestamp columns

        Example bin_specs dictionary is:

        >>> bin_specs = {'x': {'bin_width': 1, 'bin_offset': 0},
                         'y': {'bin_edges': [0, 2, 3, 4, 5, 7, 8]}}

        :param dict var_dtype: dict of datatypes of the columns to study from dataframe
                               If not provided, try to determine datatypes directy from dataframe.
        :param dict quantity: dictionary of lambda functions of how to pars certain columns

        Example quantity dictionary is:

        >>> quantity = {'y': lambda x: x}

        :param bool store_at_finalize: Store histograms in datastore at finalize(), not at
                                       execute(). Useful when looping over datasets. Default is False.
        :param drop_keys dict: dictionary used for dropping specific keys from bins dictionaries of histograms

        Example drop_keys dictionary is:

        >>> drop_keys = {'x': [1, 4, 8, 19],
                         'y': ['apple', 'pear', 'tomato'],
                         'x:y': [(1, 'apple'), (19, 'tomato')]}
        """
        # initialize Link, pass name from kwargs
        if 'name' not in kwargs:
            kwargs['name'] = 'HistogrammarFiller'
        HistogramFillerBase.__init__(self, **kwargs)

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs,
                             quantity={})

    def fill_histogram(self, idf, columns):
        """Fill input histogram with column(s) of input dataframe.

        :param idf: input data frame used for filling histogram
        :param list columns: histogram column(s)
        """
        name = ':'.join(columns)
        if name not in self._hists:
            # create an (empty) histogram of right type
            self._hists[name] = self.construct_empty_hist(columns)
        hist = self._hists[name]
        # do the actual filling
        clm = columns[0] if len(columns) == 1 else columns
        hist.fill.numpy(idf[clm])
        # remove specific keys from histogram before merging, if so requested
        hist.bins = self.drop_requested_keys(name, hist.bins)
        self._hists[name] = hist

    def construct_empty_hist(self, columns):
        """Create an (empty) histogram of right type.

        Create a multi-dim histogram by iterating through the columns in
        reverse order and passing a single-dim hist as input to the next
        column.

        :param list columns: histogram columns
        :returns: created histogram
        :rtype: histogrammar.Count
        """
        hist = hg.Count()

        # create a multi-dim histogram by iterating through the columns in reverse order
        # and passing a single-dim hist as input to the next column
        for col in reversed(columns):
            # histogram type depends on the data type
            dt = np.dtype(self.var_dtype[col])

            # processing function, e.g. only accept boolians during filling
            f = self.quantity.get(col, hf.QUANTITY[dt.type])
            if len(columns) == 1:
                # df[col] is a pd.series
                quant = lambda x, fnc=f: fnc(x)  # noqa
            else:
                # df[columns] is a pd.Dataframe
                # fix column to col
                quant = lambda x, fnc=f, clm=col: fnc(x[clm])  # noqa

            is_number = isinstance(dt.type(), np.number)
            is_timestamp = isinstance(dt.type(), np.datetime64)

            if is_number or is_timestamp:
                # numbers and timestamps are put in a sparse binned histogram
                bs = self.bin_specs.get(col, self._unit_bin_specs if is_number else self._unit_timestamp_specs)
                hist = hg.SparselyBin(binWidth=bs['bin_width'], origin=bs['bin_offset'], quantity=quant, value=hist)
            else:
                # string and boolians are treated as categories
                hist = hg.Categorize(quantity=quant, value=hist)

        # FIXME stick data types and number of dimension to histogram
        dta = [self.var_dtype[col] for col in columns]
        hist.datatype = dta[0] if len(columns) == 1 else dta
        hist.n_dim = len(columns)

        @property
        def n_bins(self):
            """`Get number of bins."""
            if hasattr(self, 'num'):
                return self.num
            elif hasattr(self, 'size'):
                return self.size
            else:
                raise RuntimeError('Cannot retrieve number of bins from hgr hist')

        hist.n_bins = n_bins

        return hist



# flake8: noqa
from eskapade import helpers
from eskapade.core.definitions import StatusCode
from eskapade.core.process_services import ConfigObject
from eskapade.core.process_services import DataStore
from eskapade.core.element import Chain, Link
from eskapade.core.process_manager import process_manager
from eskapade.exceptions import *
from eskapade.version import version as __version__



"""Project: Eskapade - A python-based package for data analysis.

Class: LinePrinter

Created: 2017/02/21

Description:
    Simple algorithm to pick up lines and reprint them.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager


class LinePrinter(Link):
    """LinePrinter picks up lines from the datastore and prints them."""

    def __init__(self, **kwargs):
        """Set up the configuration of link LinePrinter.

        :param str name: name of link
        :param str read_key: key of input data to read from data store
        """
        # initialize Link, pass name from kwargs
        Link.__init__(self, kwargs.pop('name', 'LinePrinter'))

        # process and register all relevant kwargs. kwargs are added as attributes of the link.
        # second arg is default value for an attribute. key is popped from kwargs.
        self._process_kwargs(kwargs, read_key=None)

        # check residual kwargs. exit if any present.
        self.check_extra_kwargs(kwargs)

    def initialize(self):
        """Initialize the link."""
        assert isinstance(self.read_key, str) and len(self.read_key), 'read_key not set properly.'

        return StatusCode.Success

    def execute(self):
        """Execute the link.

        No output is printed except for lines that are passed on,
        such that the output lines can be picked up again by another parser.
        """
        ds = process_manager.service(DataStore)

        # just print the lines!
        lines = ds[self.read_key]
        assert isinstance(lines, list) and len(lines), 'lines is not a (filled) list.'

        for line in lines:
            print(line)

        return StatusCode.Success



"""Project: Eskapade - A python-based package for data analysis.

Macro: esk105_datastore_pickling

Created: 2017/02/20

Description:
    Macro illustrates how to persist the datastore
    and configuration object after each chain.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

from eskapade import ConfigObject
from eskapade import process_manager, resources
from eskapade.logger import Logger

logger = Logger()

#########################################################################################
# --- Analysis values, settings, helper functions, configuration flags.

# turning on this flag, the datastore and configuration are not written out to disk
# at the end of the program.

settings = process_manager.service(ConfigObject)
settings['storeResultsEachChain'] = True

msg = r"""

The global flag settings['storeResultsEachChain'] (default = False)
controls persistence of the run-process services after the execution of
a chain.  By default, these objects are only stored after the last
chain.

From the command line, this flag is set with option --store-all.
Alternatively, the option --store-one="chain_name" can be used to store
the results of a single chain.  The with the option --store-none, no
results are stored.
"""
logger.info(msg)

#########################################################################################
# --- now parse the follow-up macro


# the flag doNotStoreResults is picked up when parsing the following macro
process_manager.execute_macro(resources.tutorial('esk105_datastore_pickling.py'))



"""Project: Eskapade - A python-based package for data analysis.

Class: DsToDs

Created: 2016/11/08

Description:
    Algorithm to move, copy, or remove an object in the datastore.

Authors:
    KPMG Advanced Analytics & Big Data team, Amstelveen, The Netherlands

Redistribution and use in source and binary forms, with or without
modification, are permitted according to the terms listed in the file
LICENSE.
"""

import copy

import pandas as pd

from eskapade import DataStore
from eskapade import Link
from eskapade import StatusCode
from eskapade import process_manager


class DsToDs(Link):
    """Link to move, copy, or remove an object in the datastore."""

    def __init__(self, **kwargs):
        """
        Initialize link instance.

        :param str name: name of link
        :param str read_key: key of data to read from data store
        :param str store_key: key of data to store in data store
        :param bool move: move read_key item to store_key. Default is true.
        :param bool copy: if True the read_key key, value pair will not be deleted. Default is false.
        :param bool remove: if True the item corresponding to read_key key will be deleted. Default is false.
        :param dict columnsToAdd: if the object is a pandas.DataFrame columns to add to the pandas.DataFrame.
            key = column name, value = column
        """
        Link.__init__(self, kwargs.pop('name', 'DsToDs'))

        # process keyword arguments
        self._process_kwargs(kwargs, read_key='', store_key='', columnsToAdd=None, move=True, copy=False, remove=False)
        self.check_extra_kwargs(kwargs)

    def initialize(self):
        """Initialize the link."""
        assert isinstance(self.read_key, str) and self.read_key, 'read_key not set.'
        if not self.remove:
            assert isinstance(self.store_key, str) and self.store_key, 'store_key not set.'

        return StatusCode.Success

    def execute(self):
        """Execute the link."""
        ds = process_manager.service(DataStore)

        if self.read_key not in ds:
            self.logger.warning('read_key <{key}> not in DataStore. Return.', key=self.read_key)
            return StatusCode.Recoverable

        # if the object is a pandas.DataFrame columns can be added to the dataframe
        df = ds[self.read_key]
        if isinstance(df, pd.DataFrame) and self.columnsToAdd is not None:
            for k, v in self.columnsToAdd.items():
                df[k] = v

        # copy, remove, or move item. default is move
        if self.copy:
            ds[self.store_key] = copy.deepcopy(ds[self.read_key])
        elif self.remove:
            ds.pop(self.read_key)
        elif self.move:
            ds[self.store_key] = ds.pop(self.read_key)

        return StatusCode.Success
